"""
A module containing detection-related plotting functions.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import scipy.stats
from typing_extensions import Literal

from .._helper import export, verify_arraylike, verify_bool, verify_isinstance, verify_literal, verify_scalar
from ._rc_params import RC_PARAMS


@export
def p_d(
    x: npt.ArrayLike,
    p_d: npt.ArrayLike,
    ax: plt.Axes | None = None,
    x_label: Literal["snr", "enr"] = "snr",
    **kwargs,
):
    r"""
    Plots the probability of detection $P_d$ as a function of received SNR or ENR.

    Arguments:
        x: The SNR or ENR in dB.
        p_d: The probability of detection $P_d$.
        ax: The axis to plot on. If `None`, the current axis is used.
        x_label: The x-axis label to use.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

    Group:
        plot-detection
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    p_d = verify_arraylike(p_d, float=True, ndim=1)
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(x_label, ["snr", "enr"])

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        default_kwargs = {}
        kwargs = {**default_kwargs, **kwargs}

        ax.plot(x, p_d, **kwargs)

        if "label" in kwargs:
            ax.legend()

        if x_label == "snr":
            ax.set_xlabel(r"Signal-to-noise ratio (dB), $S/\sigma^2$")
        elif x_label == "enr":
            ax.set_xlabel(r"Energy-to-noise ratio (dB), $\mathcal{E}/\sigma^2$")

        ax.set_ylabel("Probability of detection, $P_d$")
        ax.set_title("Detection performance")


@export
def roc(
    p_fa: npt.ArrayLike,
    p_d: npt.ArrayLike,
    ax: plt.Axes | None = None,
    type: Literal["linear", "semilogx", "semilogy", "loglog"] = "semilogx",
    **kwargs,
):
    r"""
    Plots the receiver operating characteristic (ROC) curve as a function of $P_{fa}$.

    Arguments:
        p_fa: The probability of false alarm $P_{fa}$.
        p_d: The probability of detection $P_d$.
        ax: The axis to plot on. If `None`, the current axis is used.
        type: The type of plot to generate.
        kwargs: Additional keyword arguments to pass to the plotting function defined by `type`.

    Group:
        plot-detection
    """
    p_fa = verify_arraylike(p_fa, float=True, ndim=1)
    p_d = verify_arraylike(p_d, float=True, ndim=1)
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(type, ["linear", "semilogx", "semilogy", "loglog"])

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        default_kwargs = {}
        kwargs = {**default_kwargs, **kwargs}

        if type == "linear":
            ax.plot(p_fa, p_d, **kwargs)
        elif type == "semilogx":
            ax.semilogx(p_fa, p_d, **kwargs)
        elif type == "semilogy":
            ax.semilogy(p_fa, p_d, **kwargs)
        elif type == "loglog":
            ax.loglog(p_fa, p_d, **kwargs)

        if "label" in kwargs:
            ax.legend()

        ax.set_xlabel("Probability of false alarm, $P_{fa}$")
        ax.set_ylabel("Probability of detection, $P_d$")
        ax.set_title("Receiver operating characteristic (ROC) curve")


@export
def detector_pdfs(
    h0: scipy.stats.rv_continuous | None = None,
    h1: scipy.stats.rv_continuous | None = None,
    threshold: float | None = None,
    shade: bool = True,
    annotate: bool = True,
    x: npt.NDArray[np.float64] | None = None,
    points: int = 1001,
    p_h0: float = 1e-6,
    p_h1: float = 1e-3,
    ax: plt.Axes | None = None,
    **kwargs,
):
    r"""
    Plots the probability density functions (PDFs) of the detector under $\mathcal{H}_0$ and $\mathcal{H}_1$.

    Arguments:
        h0: The statistical distribution under $\mathcal{H}_0$.
        h1: The statistical distribution under $\mathcal{H}_1$.
        threshold: The detection threshold $\gamma$.
        shade: Indicates whether to shade the tails of the PDFs.
        annotate: Indicates whether to annotate the plot with the probabilities of false alarm and detection.
        x: The x-axis values to use for the plot. If not provided, it will be generated automatically.
        points: The number of points to use for the x-axis.
        p_h0: The probability of the $\mathcal{H}_0$ tails to plot. The smaller the value, the longer the x-axis.
        p_h1: The probability of the $\mathcal{H}_1$ tails to plot. The smaller the value, the longer the x-axis.
        ax: The axis to plot on. If `None`, the current axis is used.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

    See Also:
        sdr.h0, sdr.h1, sdr.threshold

    Example:
        .. ipython:: python

            snr = 5  # Signal-to-noise ratio in dB
            sigma2 = 1  # Noise variance
            p_fa = 1e-1  # Probability of false alarm

        .. ipython:: python

            detector = "linear"; \
            h0 = sdr.h0(sigma2, detector); \
            h1 = sdr.h1(snr, sigma2, detector); \
            threshold = sdr.threshold(p_fa, sigma2, detector)

            @savefig sdr_plot_detector_pdfs_1.png
            plt.figure(); \
            sdr.plot.detector_pdfs(h0, h1, threshold); \
            plt.title("Linear Detector: Probability density functions");

    Group:
        plot-detection
    """
    # verify_isinstance(h0, scipy.stats.rv_continuous, optional=True)  # TODO: Need a better stats check
    # verify_isinstance(h1, scipy.stats.rv_continuous, optional=True)  # TODO: Need a better stats check
    verify_scalar(threshold, optional=True, float=True)
    verify_bool(shade)
    verify_bool(annotate)
    verify_arraylike(x, optional=True, float=True, ndim=1)
    verify_scalar(points, int=True, positive=True)
    verify_scalar(p_h0, float=True, positive=True)
    verify_scalar(p_h1, float=True, positive=True)
    verify_isinstance(ax, plt.Axes, optional=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        default_kwargs = {}
        kwargs = {**default_kwargs, **kwargs}

        if x is None:
            x_min = []
            if h0 is not None:
                x_min.append(h0.ppf(p_h0))
            if h1 is not None:
                x_min.append(h1.ppf(p_h1))
            if threshold is not None:
                x_min.append(threshold)
            x_min = np.nanmin(x_min)

            x_max = []
            if h0 is not None:
                x_max.append(h0.isf(p_h0))
            if h1 is not None:
                x_max.append(h1.isf(p_h1))
            if threshold is not None:
                x_max.append(threshold)
            x_max = np.nanmax(x_max)

            x = np.linspace(x_min, x_max, points)

        if h0 is not None:
            ax.plot(x, h0.pdf(x), label=r"$\mathcal{H}_0$: Noise", **kwargs)
            if shade and threshold:
                h0_color = ax.lines[-1].get_color()
                ax.fill_between(x, 0, h0.pdf(x), where=(x >= threshold), interpolate=True, color=h0_color, alpha=0.1)
        if h1 is not None:
            ax.plot(x, h1.pdf(x), label=r"$\mathcal{H}_1$: Signal + Noise", **kwargs)
            if shade and threshold:
                h1_color = ax.lines[-1].get_color()
                ax.fill_between(x, 0, h1.pdf(x), where=(x >= threshold), interpolate=True, color=h1_color, alpha=0.1)
        if threshold is not None:
            ax.axvline(threshold, color="k", linestyle="--", label="Threshold")

        if annotate:
            if h0 is not None and threshold is not None:
                p_fa = h0.sf(threshold)
                threshold_half = h0.isf(p_fa / 2)
                ax.text(
                    threshold_half,
                    h0.pdf(threshold_half) / 2,
                    rf"$P_{{fa}} = $ {p_fa:1.2e}",
                    color=h0_color,
                    ha="center",
                    va="center",
                )

            if h1 is not None and threshold is not None:
                p_d = h1.sf(threshold)
                threshold_half = h1.isf(p_d / 2)
                ax.text(
                    threshold_half,
                    h1.pdf(threshold_half) / 2,
                    rf"$P_{{d}} = $ {p_d:1.2e}",
                    color=h1_color,
                    ha="center",
                    va="center",
                )

        ax.legend()
        ax.set_xlabel("Test statistic, $T(x)$")
        ax.set_ylabel("Probability density")
        ax.set_title("Detector probability density functions (PDFs)")



"""
A module containing various discrete-time pulse shapes.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
from typing_extensions import Literal

from .._helper import export, verify_literal, verify_scalar


@export
def rectangular(
    sps: int,
    span: int = 1,
    norm: Literal["power", "energy", "passband"] = "energy",
) -> npt.NDArray[np.float64]:
    r"""
    Returns a rectangular pulse shape.

    Arguments:
        sps: The number of samples per symbol.
        span: The length of the filter in symbols. The length of the filter is `span * sps` samples,
            but only the center `sps` samples are non-zero. The only reason for `span` to be larger than 1 is to
            add delay to the filter.
        norm: Indicates how to normalize the pulse shape.

            - `"power"`: The pulse shape is normalized so that the maximum power is 1.
            - `"energy"`: The pulse shape is normalized so that the total energy is 1.
            - `"passband"`: The pulse shape is normalized so that the passband gain is 1.

    Returns:
        The rectangular pulse shape.

    Examples:
        .. ipython:: python

            h_rect = sdr.rectangular(10)

            @savefig sdr_rectangular_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h_rect);

            @savefig sdr_rectangular_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_rect);

        See the :ref:`pulse-shapes` example.

    Group:
        modulation-pulse-shaping
    """
    verify_scalar(sps, int=True, positive=True)
    verify_scalar(span, int=True, positive=True)
    verify_literal(norm, ["power", "energy", "passband"])

    length = span * sps
    h = np.zeros(length, dtype=float)
    idx = (length - sps) // 2
    h[idx : idx + sps] = 1

    h = _normalize(h, norm)

    return h


@export
def half_sine(
    sps: int,
    span: int = 1,
    norm: Literal["power", "energy", "passband"] = "energy",
) -> npt.NDArray[np.float64]:
    r"""
    Returns a half-sine pulse shape.

    Arguments:
        sps: The number of samples per symbol.
        span: The length of the filter in symbols. The length of the filter is `span * sps` samples,
            but only the center `sps` samples are non-zero. The only reason for `span` to be larger than 1 is to
            add delay to the filter.
        norm: Indicates how to normalize the pulse shape.

            - `"power"`: The pulse shape is normalized so that the maximum power is 1.
            - `"energy"`: The pulse shape is normalized so that the total energy is 1.
            - `"passband"`: The pulse shape is normalized so that the passband gain is 1.

    Returns:
        The half-sine pulse shape.

    Examples:
        .. ipython:: python

            h_half_sine = sdr.half_sine(10)

            @savefig sdr_half_sine_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h_half_sine);

            @savefig sdr_half_sine_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_half_sine);

        See the :ref:`pulse-shapes` example.

    Group:
        modulation-pulse-shaping
    """
    verify_scalar(sps, int=True, positive=True)
    verify_scalar(span, int=True, positive=True)
    verify_literal(norm, ["power", "energy", "passband"])

    length = span * sps
    h = np.zeros(length, dtype=float)
    idx = (length - sps) // 2
    h[idx : idx + sps] = np.sin(np.pi * np.arange(sps) / sps)

    h = _normalize(h, norm)

    return h


@export
def gaussian(
    time_bandwidth: float,
    span: int,
    sps: int,
    norm: Literal["power", "energy", "passband"] = "passband",
) -> npt.NDArray[np.float64]:
    r"""
    Returns a Gaussian pulse shape.

    Arguments:
        time_bandwidth: The time-bandwidth product $B T_{sym}$ of the filter, where $B$ is the one-sided
            3-dB bandwidth in Hz and $T_{sym}$ is the symbol time in seconds. The time-bandwidth product
            can also be thought of as the fractional bandwidth $B / f_{sym}$. Smaller values produce
            wider pulses.
        span: The length of the filter in symbols. The length of the filter is `span * sps + 1` samples.
            The filter order `span * sps` must be even.
        sps: The number of samples per symbol.
        norm: Indicates how to normalize the pulse shape.

            - `"power"`: The pulse shape is normalized so that the maximum power is 1.
            - `"energy"`: The pulse shape is normalized so that the total energy is 1.
            - `"passband"`: The pulse shape is normalized so that the passband gain is 1.

    Returns:
        The Gaussian pulse shape.

    Notes:
        The Gaussian pulse shape has a transfer function of

        $$H(f) = \exp(-\alpha^2 f^2) .$$

        The parameter $\alpha$ is related to the 3-dB bandwidth $B$ by

        $$\alpha = \sqrt{\frac{\ln 2}{2}} \frac{T_{sym}}{B T_{sym}} = \sqrt{\frac{\ln 2}{2}} \frac{1}{B}.$$

        The impulse response is defined as

        $$h(t) = \frac{\sqrt{\pi}}{\alpha} \exp\left[-\left(\frac{\pi}{\alpha} t\right)^2 \right]. $$

    References:
        - https://www.mathworks.com/help/signal/ref/gaussdesign.html
        - `Appendix B: Gaussian Pulse-Shaping Filter.
          <https://onlinelibrary.wiley.com/doi/pdf/10.1002/9780470041956.app2>`_

    Examples:
        .. ipython:: python

            h_0p1 = sdr.gaussian(0.1, 5, 10); \
            h_0p2 = sdr.gaussian(0.2, 5, 10); \
            h_0p3 = sdr.gaussian(0.3, 5, 10);

            @savefig sdr_gaussian_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h_0p1, label=r"$B T_{sym} = 0.1$"); \
            sdr.plot.impulse_response(h_0p2, label=r"$B T_{sym} = 0.2$"); \
            sdr.plot.impulse_response(h_0p3, label=r"$B T_{sym} = 0.3$")

            @savefig sdr_gaussian_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_0p1, label=r"$B T_{sym} = 0.1$"); \
            sdr.plot.magnitude_response(h_0p2, label=r"$B T_{sym} = 0.2$"); \
            sdr.plot.magnitude_response(h_0p3, label=r"$B T_{sym} = 0.3$")

        See the :ref:`pulse-shapes` example.

    Group:
        modulation-pulse-shaping
    """
    verify_scalar(time_bandwidth, float=True, positive=True)
    verify_scalar(span, int=True, positive=True)
    verify_scalar(sps, int=True, inclusive_min=2)
    verify_scalar(span * sps, even=True)
    verify_literal(norm, ["power", "energy", "passband"])

    t = np.arange(-(span * sps) // 2, (span * sps) // 2 + 1, dtype=float)
    t /= sps

    # Equation B.2
    Ts = 1
    alpha = np.sqrt(np.log(2) / 2) * Ts / time_bandwidth

    # Equation B.3
    h = np.sqrt(np.pi) / alpha * np.exp(-((np.pi * t / alpha) ** 2))

    h = _normalize(h, norm)

    return h


@export
def raised_cosine(
    alpha: float,
    span: int,
    sps: int,
    norm: Literal["power", "energy", "passband"] = "energy",
) -> npt.NDArray[np.float64]:
    r"""
    Returns a raised cosine (RC) pulse shape.

    Arguments:
        alpha: The excess bandwidth $0 \le \alpha \le 1$ of the filter.
        span: The length of the filter in symbols. The length of the filter is `span * sps + 1` samples.
            The filter order `span * sps` must be even.
        sps: The number of samples per symbol.
        norm: Indicates how to normalize the pulse shape.

            - `"power"`: The pulse shape is normalized so that the maximum power is 1.
            - `"energy"`: The pulse shape is normalized so that the total energy is 1.
            - `"passband"`: The pulse shape is normalized so that the passband gain is 1.

    Returns:
        The raised cosine pulse shape.

    Notes:
        The raised cosine pulse shape has a transfer function of

        $$
        H(f) =
        \begin{cases}
        T_{sym},
        & \displaystyle 0 \le \left| f \right| \le \frac{1 - \alpha}{2 T_{sym}} \\
        \displaystyle \frac{T_{sym}}{2} \left[1 + \cos\left(\frac{\pi T_{sym}}{\alpha}\left(\left| f \right| - \frac{1 - \alpha}{2 T_{sym}}\right)\right)\right],
        & \displaystyle \frac{1 - \alpha}{2 T_{sym}} \le \left| f \right| \le \frac{1 + \alpha}{2 T_{sym}} \\
        0,
        & \displaystyle \left| f \right| \ge \frac{1 + \alpha}{2 T_{sym}}
        \end{cases}
        $$

        The impulse response is defined as

        $$
        h(t) =
        \frac
        {\displaystyle \sin\left(\frac{\pi t}{T_{sym}}\right) \cos\left(\frac{\pi \alpha t}{T_{sym}}\right)}
        {\displaystyle \frac{\pi t}{T_{sym}} \left[1 - \left(\frac{2 \alpha t}{T_{sym}} \right)^2 \right]}
        $$

    References:
        - Michael Rice, *Digital Communications: A Discrete Time Approach*, Appendix A.

    Examples:
        The excess bandwidth $\alpha$ controls bandwidth of the filter. A smaller $\alpha$ results in a
        narrower bandwidth at the expense of higher sidelobes.

        .. ipython:: python

            h_0p1 = sdr.raised_cosine(0.1, 8, 10); \
            h_0p5 = sdr.raised_cosine(0.5, 8, 10); \
            h_0p9 = sdr.raised_cosine(0.9, 8, 10)

            @savefig sdr_raised_cosine_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h_0p1, label=r"$\alpha = 0.1$"); \
            sdr.plot.impulse_response(h_0p5, label=r"$\alpha = 0.5$"); \
            sdr.plot.impulse_response(h_0p9, label=r"$\alpha = 0.9$")

            @savefig sdr_raised_cosine_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_0p1, label=r"$\alpha = 0.1$"); \
            sdr.plot.magnitude_response(h_0p5, label=r"$\alpha = 0.5$"); \
            sdr.plot.magnitude_response(h_0p9, label=r"$\alpha = 0.9$")

        The span of the filter affects the stopband attenuation. A longer span results in greater stopband
        attenuation and lower sidelobes.

        .. ipython:: python

            h_4 = sdr.raised_cosine(0.1, 4, 10); \
            h_8 = sdr.raised_cosine(0.1, 8, 10); \
            h_16 = sdr.raised_cosine(0.1, 16, 10)

            @savefig sdr_raised_cosine_3.png
            plt.figure(); \
            sdr.plot.impulse_response(h_4, label="span = 4"); \
            sdr.plot.impulse_response(h_8, label="span = 8"); \
            sdr.plot.impulse_response(h_16, label="span = 16")

            @savefig sdr_raised_cosine_4.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_4, label="span = 4"); \
            sdr.plot.magnitude_response(h_8, label="span = 8"); \
            sdr.plot.magnitude_response(h_16, label="span = 16")

        See the :ref:`pulse-shapes` example.

    Group:
        modulation-pulse-shaping
    """
    verify_scalar(alpha, float=True, inclusive_min=0, inclusive_max=1)
    verify_scalar(span, int=True, positive=True)
    verify_scalar(sps, int=True, inclusive_min=2)
    verify_scalar(span * sps, even=True)
    verify_literal(norm, ["power", "energy", "passband"])

    t = np.arange(-(span * sps) // 2, (span * sps) // 2 + 1, dtype=float)
    Ts = sps

    # Handle special cases where the denominator is zero
    t[t == 0] += 1e-16
    if alpha > 0:
        t1 = Ts / (2 * alpha)
        t[t == -t1] -= 1e-8
        t[t == t1] += 1e-8

    # Equation A-27
    A = np.sin(np.pi * t / Ts) / (np.pi * t / Ts)
    B = np.cos(np.pi * alpha * t / Ts) / (1 - (2 * alpha * t / Ts) ** 2)
    h = A * B

    h = _normalize(h, norm)

    return h


@export
def root_raised_cosine(
    alpha: float,
    span: int,
    sps: int,
    norm: Literal["power", "energy", "passband"] = "energy",
) -> npt.NDArray[np.float64]:
    r"""
    Returns a square root raised cosine (SRRC) pulse shape.

    Arguments:
        alpha: The excess bandwidth $0 \le \alpha \le 1$ of the filter.
        span: The length of the filter in symbols. The length of the filter is `span * sps + 1` samples.
            The filter order `span * sps` must be even.
        sps: The number of samples per symbol.
        norm: Indicates how to normalize the pulse shape.

            - `"power"`: The pulse shape is normalized so that the maximum power is 1.
            - `"energy"`: The pulse shape is normalized so that the total energy is 1.
            - `"passband"`: The pulse shape is normalized so that the passband gain is 1.

    Returns:
        The square-root raised cosine pulse shape.

    Notes:
        The square root raised cosine pulse shape has a transfer function of

        $$
        H(f) =
        \begin{cases}
        \displaystyle \sqrt{T_{sym}},
        & \displaystyle 0 \le \left| f \right| \le \frac{1 - \alpha}{2 T_{sym}} \\
        \displaystyle \sqrt{\frac{T_{sym}}{2} \left[1 + \cos\left(\frac{\pi T_{sym}}{\alpha}\left(\left| f \right| - \frac{1 - \alpha}{2 T_{sym}}\right)\right)\right]},
        & \displaystyle \frac{1 - \alpha}{2 T_{sym}} \le \left| f \right| \le \frac{1 + \alpha}{2 T_{sym}} \\
        0,
        & \displaystyle \left| f \right| \ge \frac{1 + \alpha}{2 T_{sym}}
        \end{cases}
        $$

        The impulse response is defined as

        $$
        h(t) =
        \frac{1}{\sqrt{T_{sym}}}
        \frac
        {\displaystyle \sin\left(\frac{\pi (1 - \alpha) t}{T_{sym}}\right) + \frac{4 \alpha t}{T_{sym}}\cos\left( \frac{\pi (1 + \alpha) t}{T_{sym}}\right)}
        {\displaystyle \frac{\pi t}{T_{sym}} \left[1 - \left(\frac{4 \alpha t}{T_{sym}} \right)^2 \right]}
        $$

    References:
        - Michael Rice, *Digital Communications: A Discrete Time Approach*, Appendix A.

    Examples:
        The excess bandwidth $\alpha$ controls bandwidth of the filter. A smaller $\alpha$ results in a
        narrower bandwidth at the expense of higher sidelobes.

        .. ipython:: python

            h_0p1 = sdr.root_raised_cosine(0.1, 8, 10); \
            h_0p5 = sdr.root_raised_cosine(0.5, 8, 10); \
            h_0p9 = sdr.root_raised_cosine(0.9, 8, 10)

            @savefig sdr_root_raised_cosine_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h_0p1, label=r"$\alpha = 0.1$"); \
            sdr.plot.impulse_response(h_0p5, label=r"$\alpha = 0.5$"); \
            sdr.plot.impulse_response(h_0p9, label=r"$\alpha = 0.9$")

            @savefig sdr_root_raised_cosine_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_0p1, label=r"$\alpha = 0.1$"); \
            sdr.plot.magnitude_response(h_0p5, label=r"$\alpha = 0.5$"); \
            sdr.plot.magnitude_response(h_0p9, label=r"$\alpha = 0.9$")

        The span of the filter affects the stopband attenuation. A longer span results in greater stopband
        attenuation and lower sidelobes.

        .. ipython:: python

            h_4 = sdr.root_raised_cosine(0.1, 4, 10); \
            h_8 = sdr.root_raised_cosine(0.1, 8, 10); \
            h_16 = sdr.root_raised_cosine(0.1, 16, 10)

            @savefig sdr_root_raised_cosine_3.png
            plt.figure(); \
            sdr.plot.impulse_response(h_4, label="span = 4"); \
            sdr.plot.impulse_response(h_8, label="span = 8"); \
            sdr.plot.impulse_response(h_16, label="span = 16")

            @savefig sdr_root_raised_cosine_4.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_4, label="span = 4"); \
            sdr.plot.magnitude_response(h_8, label="span = 8"); \
            sdr.plot.magnitude_response(h_16, label="span = 16")

        See the :ref:`pulse-shapes` example.

    Group:
        modulation-pulse-shaping
    """
    verify_scalar(alpha, float=True, inclusive_min=0, inclusive_max=1)
    verify_scalar(span, int=True, positive=True)
    verify_scalar(sps, int=True, inclusive_min=2)
    verify_scalar(span * sps, even=True)
    verify_literal(norm, ["power", "energy", "passband"])

    t = np.arange(-(span * sps) // 2, (span * sps) // 2 + 1, dtype=float)
    Ts = sps  # Symbol duration (in samples)

    # Handle special cases where the denominator is zero
    t[t == 0] += 1e-16
    if alpha > 0:
        t1 = Ts / (4 * alpha)
        t[t == -t1] -= 1e-8
        t[t == t1] += 1e-8

    # Equation A-30
    A = np.sin(np.pi * (1 - alpha) * t / Ts)
    B = 4 * alpha * t / Ts * np.cos(np.pi * (1 + alpha) * t / Ts)
    C = np.pi * t / Ts * (1 - (4 * alpha * t / Ts) ** 2)
    h = (A + B) / C
    h /= 1 / np.sqrt(Ts)

    h = _normalize(h, norm)

    return h


def _normalize(h: npt.NDArray[np.float64], norm: Literal["power", "energy", "passband"]) -> npt.NDArray[np.float64]:
    if norm == "power":
        h /= np.sqrt(np.max(np.abs(h) ** 2))
    elif norm == "energy":
        h /= np.sqrt(np.sum(np.abs(h) ** 2))
    elif norm == "passband":
        h /= np.sum(h)

    return h



"""
A module containing functions for various symbol mapping schemes.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._helper import export, verify_scalar


@export
def binary_code(length: int) -> npt.NDArray[np.int_]:
    """
    Generates a binary code of length $n = 2^m$.

    Arguments:
        length: The length $n = 2^m$ of the binary code.

    Returns:
        A binary code of length $n = 2^m$.

    Examples:
        .. ipython:: python

            sdr.binary_code(2)
            sdr.binary_code(4)
            sdr.binary_code(8)
            sdr.binary_code(16)

    Group:
        sequences-symbol-mapping
    """
    verify_scalar(length, int=True, positive=True, power_of_two=True)

    return np.arange(length)


@export
def gray_code(length: int) -> npt.NDArray[np.int_]:
    """
    Generates a Gray code of length $n = 2^m$.

    Arguments:
        length: The length $n = 2^m$ of the Gray code.

    Returns:
        A Gray code of length $n = 2^m$.

    Examples:
        .. ipython:: python

            sdr.gray_code(2)
            sdr.gray_code(4)
            sdr.gray_code(8)
            sdr.gray_code(16)

    Group:
        sequences-symbol-mapping
    """
    verify_scalar(length, int=True, positive=True, power_of_two=True)

    if length == 2:
        return np.array([0, 1])

    # Generate the Gray code with degree m - 1
    code_1 = gray_code(length // 2)

    # Generate the Gray code for degree m by concatenating the Gray code for degree m - 1
    # with itself reversed. Also, the most significant bit of the second half
    # is set to 1.
    return np.concatenate((code_1, code_1[::-1] + length // 2))



import numpy as np
import pytest

import sdr
import sdr._helper


def test_verify_positional_args():
    x, y = 1, "two"
    sdr._helper.verify_positional_args((x, y), 2)

    with pytest.raises(ValueError):
        sdr._helper.verify_positional_args((x, y), 1)


def test_verify_specified():
    x = 1
    sdr._helper.verify_specified(x)

    with pytest.raises(ValueError):
        x = None
        sdr._helper.verify_specified(None)


def test_verify_not_specified():
    x = None
    sdr._helper.verify_not_specified(x)

    with pytest.raises(ValueError):
        x = 1
        sdr._helper.verify_not_specified(x)


def test_verify_only_one_specified():
    x, y = 1, None
    sdr._helper.verify_only_one_specified(x, y)

    with pytest.raises(ValueError):
        x, y = 1, 2
        sdr._helper.verify_only_one_specified(x, y)

    with pytest.raises(ValueError):
        x, y = None, None
        sdr._helper.verify_only_one_specified(x, y)


def test_verify_at_least_one_specified():
    x, y = 1, None
    sdr._helper.verify_at_least_one_specified(x, y)

    x, y = 1, "two"
    sdr._helper.verify_at_least_one_specified(x, y)

    with pytest.raises(ValueError):
        x, y = None, None
        sdr._helper.verify_at_least_one_specified(x, y)


def test_verify_isinstance():
    x = 1
    sdr._helper.verify_isinstance(x, int)

    x = "one"
    sdr._helper.verify_isinstance(x, str)

    with pytest.raises(TypeError):
        x = 1
        sdr._helper.verify_isinstance(x, str)

    with pytest.raises(TypeError):
        x = "one"
        sdr._helper.verify_isinstance(x, int)

    with pytest.raises(TypeError):
        x = 3.14
        sdr._helper.verify_isinstance(x, (int, str))


def test_verify_scalar():
    sdr._helper.verify_scalar(None, optional=True, int=True)
    with pytest.raises(TypeError):
        sdr._helper.verify_scalar(None, int=True)

    sdr._helper.verify_scalar(1, int=True)
    with pytest.raises(TypeError):
        sdr._helper.verify_scalar(3.14, int=True)

    sdr._helper.verify_scalar(3.14, float=True)
    with pytest.raises(TypeError):
        sdr._helper.verify_scalar(3.14 + 1j, float=True)

    sdr._helper.verify_scalar(3.14 + 1j, complex=True)
    with pytest.raises(TypeError):
        sdr._helper.verify_scalar("3.14", complex=True)

    sdr._helper.verify_scalar(3.14, real=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar(3.14 + 1j, real=True)

    sdr._helper.verify_scalar(3.14 + 1j, imaginary=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar("3.14", imaginary=True)

    sdr._helper.verify_scalar(-1, negative=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar(1, negative=True)

    sdr._helper.verify_scalar(0, non_negative=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar(-1, non_negative=True)

    sdr._helper.verify_scalar(1, positive=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar(0, positive=True)

    sdr._helper.verify_scalar(2, even=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar(3, even=True)

    sdr._helper.verify_scalar(3, odd=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar(2, odd=True)

    sdr._helper.verify_scalar(8, power_of_two=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar(7, power_of_two=True)

    sdr._helper.verify_scalar(3, inclusive_min=3)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar(2, inclusive_min=3)

    sdr._helper.verify_scalar(3, inclusive_max=3)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar(4, inclusive_max=3)

    sdr._helper.verify_scalar(4, exclusive_min=3)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar(3, exclusive_min=3)

    sdr._helper.verify_scalar(3, exclusive_max=4)
    with pytest.raises(ValueError):
        sdr._helper.verify_scalar(4, exclusive_max=4)

    # TODO: Test accept NumPy and convert NumPy


def test_verify_arraylike():
    sdr._helper.verify_arraylike(None, optional=True, int=True)
    with pytest.raises(TypeError):
        sdr._helper.verify_arraylike(None, int=True)

    sdr._helper.verify_arraylike(1, int=True)
    with pytest.raises(TypeError):
        sdr._helper.verify_arraylike(3.14, int=True)

    sdr._helper.verify_arraylike(3.14, float=True)
    with pytest.raises(TypeError):
        sdr._helper.verify_arraylike(3.14 + 1j, float=True)

    sdr._helper.verify_arraylike(3.14 + 1j, complex=True)
    with pytest.raises(TypeError):
        sdr._helper.verify_arraylike("3.14", complex=True)

    sdr._helper.verify_arraylike(3.14, real=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_arraylike(3.14 + 1j, real=True)

    sdr._helper.verify_arraylike(3.14 + 1j, imaginary=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_arraylike("3.14", imaginary=True)

    sdr._helper.verify_arraylike(-1, negative=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_arraylike(1, negative=True)

    sdr._helper.verify_arraylike(0, non_negative=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_arraylike(-1, non_negative=True)

    sdr._helper.verify_arraylike(1, positive=True)
    with pytest.raises(ValueError):
        sdr._helper.verify_arraylike(0, positive=True)

    sdr._helper.verify_arraylike(3, inclusive_min=3)
    with pytest.raises(ValueError):
        sdr._helper.verify_arraylike(2, inclusive_min=3)

    sdr._helper.verify_arraylike(3, inclusive_max=3)
    with pytest.raises(ValueError):
        sdr._helper.verify_arraylike(4, inclusive_max=3)

    sdr._helper.verify_arraylike(4, exclusive_min=3)
    with pytest.raises(ValueError):
        sdr._helper.verify_arraylike(3, exclusive_min=3)

    sdr._helper.verify_arraylike(3, exclusive_max=4)
    with pytest.raises(ValueError):
        sdr._helper.verify_arraylike(4, exclusive_max=4)


def test_verify_bool():
    x = True
    sdr._helper.verify_bool(x)

    x = np.bool_(True)
    sdr._helper.verify_bool(x)

    with pytest.raises(TypeError):
        x = 1
        sdr._helper.verify_bool(x)

    with pytest.raises(TypeError):
        x = "one"
        sdr._helper.verify_bool(x)

    with pytest.raises(TypeError):
        x = 3.14
        sdr._helper.verify_bool(x)

    with pytest.raises(TypeError):
        x = np.bool_(True)
        sdr._helper.verify_bool(x, accept_numpy=False)


def test_verify_literal():
    x = 3.14
    sdr._helper.verify_literal(x, (3.14, 2.718))

    with pytest.raises(ValueError):
        x = 1
        sdr._helper.verify_literal(x, (2, 3))

    with pytest.raises(ValueError):
        x = "one"
        sdr._helper.verify_literal(x, ("two", "three"))


def test_verify_coprime():
    x, y = 3, 5
    sdr._helper.verify_coprime(x, y)

    with pytest.raises(ValueError):
        x, y = 3, 6
        sdr._helper.verify_coprime(x, y)


def test_verify_condition():
    x = 3
    sdr._helper.verify_condition(x > 2)

    with pytest.raises(ValueError):
        x = 1
        sdr._helper.verify_condition(x > 2)


def test_verify_same_shape():
    x = np.zeros((2, 3))
    y = np.zeros((2, 3))
    sdr._helper.verify_same_shape(x, y)

    with pytest.raises(ValueError):
        x = np.zeros((2, 3))
        y = np.zeros((2, 4))
        sdr._helper.verify_same_shape(x, y)


def test_convert_output():
    assert sdr._helper.convert_output(1) == 1
    assert sdr._helper.convert_output(np.array(1)) == 1
    assert sdr._helper.convert_output(np.array([1]), squeeze=True) == 1

    assert sdr._helper.convert_output(3.14) == 3.14
    assert sdr._helper.convert_output(np.array(3.14)) == 3.14
    assert sdr._helper.convert_output(np.array([3.14]), squeeze=True) == 3.14

    assert sdr._helper.convert_output(3.14 + 1j) == 3.14 + 1j
    assert sdr._helper.convert_output(np.array(3.14 + 1j)) == 3.14 + 1j
    assert sdr._helper.convert_output(np.array([3.14 + 1j]), squeeze=True) == 3.14 + 1j

    assert sdr._helper.convert_output(True) is True
    assert sdr._helper.convert_output(np.array(True)) is True
    assert sdr._helper.convert_output(np.array([True]), squeeze=True) is True

    x = np.array([1, 2, 3])
    assert np.array_equal(sdr._helper.convert_output(x), x)



"""
A module containing various energy measurement functions.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._conversion import db as to_db
from .._helper import convert_output, export, verify_arraylike


@export
def energy(
    x: npt.ArrayLike,
    axis: int | tuple[int, ...] | None = None,
    db: bool = False,
) -> npt.NDArray[np.float64]:
    r"""
    Measures the energy of a time-domain signal $x[n]$.

    $$E = \sum_{n=0}^{N-1} \left| x[n] \right|^2$$

    Arguments:
        x: The time-domain signal $x[n]$ to measure.
        axis: Axis or axes along which to compute the energy. The default is `None`, which computes the energy of
            the entire array.
        db: Indicates whether to return the result in decibels (dB).

    Returns:
        The signal energy. If `db=False`, $E$ is returned. If `db=True`, $10 \log_{10} E$ is returned.

    Group:
        measurement-energy
    """
    x = verify_arraylike(x, complex=True)

    E = np.sum(np.abs(x) ** 2, axis=axis)
    if db:
        E = to_db(E)

    return convert_output(E)



"""
A module containing a base class for continuous phase modulations.
"""

from __future__ import annotations

import abc

import numpy as np
import numpy.typing as npt
from typing_extensions import Literal

from .._filter import Decimator, Interpolator
from .._helper import export, verify_arraylike, verify_condition, verify_literal, verify_scalar
from .._nco import NCO
from .._sequence import binary_code, gray_code
from ._pulse_shapes import rectangular


@export
class CPM:
    r"""
    Implements continuous-phase modulation (CPM).

    .. nomenclature::
        :collapsible:

        - $k$: Symbol index
        - $n$: Sample index
        - $s[k]$: Decimal symbols
        - $a[k]$ Complex symbols
        - $x[n]$: Pulse-shaped complex samples
        - $\tilde{x}[n]$: Received (noisy) pulse-shaped complex samples
        - $\tilde{a}[k]$: Received (noisy) complex symbols
        - $\hat{a}[k]$: Complex symbol decisions
        - $\hat{s}[k]$: Decimal symbol decisions

    Group:
        modulation-continuous-phase
    """

    def __init__(
        self,
        order: int,
        index: float = 0.5,
        symbol_labels: Literal["bin", "gray"] | npt.ArrayLike = "bin",
        phase_offset: float = 0.0,
        sps: int = 8,
        # pulse_shape: npt.ArrayLike | Literal["rect", "rc", "srrc", "gaussian"] = "rect",
        pulse_shape: npt.ArrayLike | Literal["rect"] = "rect",
        span: int = 1,
        # alpha: float | None = None,
        # time_bandwidth: float | None = None,
    ):
        r"""
        Creates a new continuous-phase modulation object.

        Arguments:
            order: The modulation order $M = 2^k$.
            index: The modulation index $h$. The modulation index is the ratio of the frequency deviation to the
                symbol rate $h = \Delta f / f_{sym}$. The phase change per symbol is $\pi h$.
            symbol_labels: The decimal symbol labels of consecutive complex symbols.

                - `"bin"`: The symbols are binary-coded. Adjacent symbols may differ by more than one bit.
                - `"gray":` The symbols are Gray-coded. Adjacent symbols only differ by one bit.
                - `npt.ArrayLike`: An $M$-length array whose indices are the default symbol labels and whose values are
                  the new symbol labels.

            phase_offset: A phase offset $\phi$ in degrees.
            sps: The number of samples per symbol $f_s / f_{sym}$.
            pulse_shape: The pulse shape $h[n]$ of the instantaneous frequency of the signal. If a string is passed,
                the pulse shape is normalized such that the maximum value is 1.

                - `npt.ArrayLike`: A custom pulse shape. It is important that `sps` matches the design
                  of the pulse shape. See :ref:`pulse-shaping-functions`.
                - `"rect"`: Rectangular pulse shape.

            span: The span of the pulse shape in symbols. This is only used if `pulse_shape` is a string.

        See Also:
            sdr.rectangular
        """
        self._order = verify_scalar(order, int=True, inclusive_min=2, power_of_two=True)  # Modulation order
        self._bps = int(np.log2(self._order))  # Coded bits per symbol
        self._index = verify_scalar(index, float=True, non_negative=True)  # Modulation index

        if isinstance(symbol_labels, str):
            verify_literal(symbol_labels, ["bin", "gray"])
            if symbol_labels == "bin":
                self._symbol_labels = binary_code(self.order)
                self._symbol_labels_str = "bin"
            elif symbol_labels == "gray":
                self._symbol_labels = gray_code(self.order)
                self._symbol_labels_str = "gray"
        else:
            symbol_labels = verify_arraylike(symbol_labels, int=True, ndim=1, size=self.order)
            verify_condition(np.unique(symbol_labels).size == self.order)
            self._symbol_labels = symbol_labels
            self._symbol_labels_str = self._symbol_labels

        self._phase_offset = verify_scalar(phase_offset, float=True)  # Phase offset in degrees
        self._sps = verify_scalar(sps, int=True, positive=True)  # Samples per symbol
        verify_scalar(span, int=True, positive=True)

        if isinstance(pulse_shape, str):
            verify_literal(pulse_shape, ["rect"])
            if pulse_shape == "rect":
                self._pulse_shape = rectangular(self.sps, span=span, norm="passband") / 2
                # self._pulse_shape = np.ones(self.sps * span) / (self.sps * span) / 2
            # elif pulse_shape == "sine":
            #     # self._pulse_shape = half_sine(self.sps, norm="passband") / 2
            #     self._pulse_shape = 1 - np.cos(2 * np.pi * np.arange(0.5, self.sps + 0.5, 1) / self.sps)
            #     self._pulse_shape = _normalize(self._pulse_shape, norm="passband") / 2
            # elif pulse_shape == "rc":
            #     if alpha is None:
            #         alpha = 0.2
            #     self._pulse_shape = raised_cosine(alpha, span, self.sps, norm="passband") / 2
            # elif pulse_shape == "gaussian":
            #     if time_bandwidth is None:
            #         time_bandwidth = 0.3
            #     self._pulse_shape = gaussian(time_bandwidth, span, self.sps, norm="passband") / 2
        else:
            self._pulse_shape = verify_arraylike(pulse_shape, float=True, ndim=1)  # Pulse shape

        # if alpha is not None and pulse_shape not in ["rc", "srrc"]:
        #     raise ValueError("Argument 'alpha' is only valid for 'rc' and 'srrc' pulse shapes, not {pulse_shape!r}.")
        # if time_bandwidth is not None and pulse_shape not in ["gaussian"]:
        #     raise ValueError("Argument 'time_bandwidth' is only valid for 'gaussian' pulse shape, not {pulse_shape!r}.")

        self._tx_filter = Interpolator(self.sps, self.pulse_shape)  # Transmit pulse shaping filter
        self._rx_filter = Decimator(self.sps, self.pulse_shape[::-1].conj())  # Receive matched filter

        self._nco = NCO()

    def __repr__(self) -> str:
        return f"sdr.{type(self).__name__}({self.order}, {self.index}, phase_offset={self.phase_offset})"

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        string += f"\n  order: {self.order}"
        string += f"\n  index: {self.index}"
        string += f"\n  phase_offset: {self.phase_offset}"
        return string

    def modulate(self, s: npt.ArrayLike) -> npt.NDArray[np.complex128]:
        r"""
        Modulates the decimal symbols $s[k]$ into pulse-shaped complex samples $x[n]$.

        Arguments:
            s: The decimal symbols $s[k]$ to modulate, $0$ to $M-1$.

        Returns:
            The pulse-shaped complex samples $x[n]$ with :obj:`sps` samples per symbol
            and length `sps * s.size + pulse_shape.size - 1`.
        """
        s = verify_arraylike(s, int=True)  # Decimal symbols
        return self._modulate(s)

    def _modulate(self, s: npt.NDArray[np.int_]) -> npt.NDArray[np.complex128]:
        s = self._symbol_labels[s]  # Relabeled decimal symbols
        freq = self.index * (2 * s - (self.order - 1))  # Instantaneous frequency
        print(freq)
        # return f
        freq_ps = self._tx_pulse_shape(freq)  # Pulse-shaped instantaneous frequency
        print(freq_ps)
        # phase_ps = np.cumsum(freq_ps)  # Pulse-shaped instantaneous phase
        # phase_ps = np.insert(phase_ps, 0, 0)  # Start with phase 0
        # phase_ps = phase_ps[:-1]  # Trim last phase
        # # return phase_ps
        # x = np.exp(1j * np.pi / self.sps * phase_ps)  # Complex samples
        # # x = np.exp(1j * (2 * np.pi / self.sps * self.index * freq_ps + self._phase_offset))  # Complex samples
        x = self._nco(2 * np.pi * freq_ps, output="complex-exp")  # Complex samples
        return x

    def _tx_pulse_shape(self, a: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        x = self._tx_filter(a, mode="full")  # Complex samples
        return x

    def demodulate(self, x_tilde: npt.ArrayLike) -> npt.NDArray[np.int_]:
        r"""
        Demodulates the pulse-shaped complex samples into decimal symbol decisions.

        This method applies matched filtering and maximum-likelihood estimation.

        Arguments:
            x_tilde: The received pulse-shaped complex samples $\tilde{x}[n]$ to demodulate, with :obj:`sps`
                samples per symbol and length `sps * s_hat.size + pulse_shape.size - 1`.

        Returns:
            - The decimal symbol decisions $\hat{s}[k]$, $0$ to $M-1$.
        """
        x_tilde = verify_arraylike(x_tilde, complex=True)  # Complex samples
        return self._demodulate(x_tilde)

    def _demodulate(self, x_tilde: npt.NDArray[np.complex128]) -> npt.NDArray[np.int_]:
        raise NotImplementedError("Demodulation for continuous-phase modulations is not supported.")

    @abc.abstractmethod
    def ber(self, ebn0: npt.ArrayLike | None = None) -> np.ndarray:
        r"""
        Computes the bit error rate (BER) at the provided $E_b/N_0$ values.

        Arguments:
            ebn0: Bit energy $E_b$ to noise PSD $N_0$ ratio in dB.

        Returns:
            The bit error rate $P_b$.

        See Also:
            sdr.esn0_to_ebn0, sdr.snr_to_ebn0
        """
        raise NotImplementedError("Bit error rate calculation for continuous-phase modulations is not supported.")

    @abc.abstractmethod
    def ser(self, esn0: npt.ArrayLike) -> np.ndarray:
        r"""
        Computes the symbol error rate (SER) at the provided $E_s/N_0$ values.

        Arguments:
            esn0: Symbol energy $E_s$ to noise PSD $N_0$ ratio in dB.

        Returns:
            The symbol error rate $P_e$.

        See Also:
            sdr.ebn0_to_esn0, sdr.snr_to_esn0
        """
        raise NotImplementedError("Symbol error rate calculation for continuous-phase modulations is not supported.")

    @property
    def order(self) -> int:
        r"""
        The modulation order $M = 2^k$.
        """
        return self._order

    @property
    def bps(self) -> int:
        r"""
        The number of coded bits per symbol $k = \log_2 M$.
        """
        return self._bps

    @property
    def index(self) -> float:
        r"""
        The modulation index $h$.

        The modulation index is the ratio of the frequency deviation to the symbol rate $h = \Delta f / f_{sym}$.
        The phase change per symbol is $\pi h$.
        """
        return self._index

    @property
    def phase_offset(self) -> float:
        r"""
        The phase offset $\phi$ in degrees.
        """
        return self._phase_offset

    @property
    def sps(self) -> int:
        r"""
        The number of samples per symbol $f_s / f_{sym}$.
        """
        return self._sps

    @property
    def pulse_shape(self) -> np.ndarray:
        r"""
        The pulse shape $h[n]$ of the instantaneous frequency of the signal.
        """
        return self._pulse_shape

    @property
    def tx_filter(self) -> Interpolator:
        r"""
        The transmit interpolating pulse shaping filter. The filter coefficients are the pulse shape $h[n]$.
        """
        return self._tx_filter

    @property
    def rx_filter(self) -> Decimator:
        r"""
        The receive decimating matched filter. The filter coefficients are matched to the pulse shape $h[-n]^*$.
        """
        return self._rx_filter



"""
A module that contains various conversion functions.
"""

from __future__ import annotations

import warnings

import numpy as np
import numpy.typing as npt
from typing_extensions import Literal

from ._helper import convert_output, export, verify_arraylike, verify_literal

##############################################################################
# Decibels
##############################################################################


@export
def db(
    x: npt.ArrayLike,
    type: Literal["value", "power", "voltage"] = "value",
) -> npt.NDArray[np.float64]:
    r"""
    Converts from linear units to decibels.

    Arguments:
        x: The input value or signal.
        type: The type of input value or signal.

            - `"value"`: The input value/signal is any value.

            $$x_{\text{dB}} = 10 \log_{10} x_{\text{linear}}$$

            - `"power"`: The input value/signal is a power measurement.

            $$P_{\text{dB}} = 10 \log_{10} P_{\text{linear}}$$

            - `"voltage"`: The input value/signal is a voltage measurement.

            $$V_{\text{dB}} = 20 \log_{10} V_{\text{linear}}$$

    Returns:
        The value or signal in dB.

    Examples:
        Convert 50 MHz to 77 dB-Hz.

        .. ipython:: python

            sdr.db(50e6)

        Convert 100 mW to 20 dBm.

        .. ipython:: python

            sdr.db(100, type="power")

        Convert 2 V to 6 dBV.

        .. ipython:: python

            sdr.db(2, type="voltage")

    Group:
        conversions-decibels
    """
    x = verify_arraylike(x, non_negative=True)
    verify_literal(type, ["value", "power", "voltage"])

    with np.errstate(divide="ignore"):
        # Ignore divide by zero warning -- we're okay with -inf
        if type == "voltage":
            x_db = 20 * np.log10(x)
        else:
            x_db = 10 * np.log10(x)

    return convert_output(x_db)


@export
def linear(
    x: npt.ArrayLike,
    type: Literal["value", "power", "voltage"] = "value",
) -> npt.NDArray[np.float64]:
    r"""
    Converts from decibels to linear units.

    Arguments:
        x: The input value or signal in dB.
        type: The type of output value or signal.

            - `"value"`: The output value/signal is any value.

            $$x_{\text{linear}} = 10^{\frac{x_{\text{dB}}}{10}}$$

            - `"power"`: The output value/signal is a power measurement.

            $$P_{\text{linear}} = 10^{\frac{P_{\text{dB}}}{10}}$$

            - `"voltage"`: The output value/signal is a voltage measurement.

            $$V_{\text{linear}} = 10^{\frac{V_{\text{dB}}}{20}}$$

    Returns:
        The value or signal in linear units.

    Examples:
        Convert 77 dB-Hz to 50 MHz.

        .. ipython:: python

            sdr.linear(77)

        Convert 20 dBm to 100 mW.

        .. ipython:: python

            sdr.linear(20, type="power")

        Convert 6 dBV to 2 V.

        .. ipython:: python

            sdr.linear(6, type="voltage")

    Group:
        conversions-decibels
    """
    x = verify_arraylike(x)
    verify_literal(type, ["value", "power", "voltage"])

    with warnings.catch_warnings():
        # Ignore scalar power overflow warning -- we're okay with inf
        warnings.simplefilter("ignore", RuntimeWarning)
        if type == "voltage":
            x_linear = 10 ** (x / 20)
        else:
            x_linear = 10 ** (x / 10)

    return convert_output(x_linear)


##############################################################################
# From Eb/N0
##############################################################################


@export
def ebn0_to_esn0(
    ebn0: npt.ArrayLike,
    bps: npt.ArrayLike,
    rate: npt.ArrayLike = 1.0,
) -> npt.NDArray[np.float64]:
    r"""
    Converts from $E_b/N_0$ to $E_s/N_0$.

    $$
    \frac{E_s}{N_0} = \frac{E_b}{N_0} \frac{k}{n} \log_2 M
    $$

    Arguments:
        ebn0: Bit energy $E_b$ to noise PSD $N_0$ ratio in dB.
        bps: Coded bits per symbol $\log_2 M$, where $M$ is the modulation order.
        rate: Code rate $r = k/n$, where $k$ is the number of information bits and $n$ is the
            number of coded bits.

    Returns:
        The symbol energy $E_s$ to noise PSD $N_0$ ratio in dB.

    Examples:
        Convert from $E_b/N_0 = 5$ dB to $E_s/N_0$ for a 4-QAM signal with $r = 2/3$.

        .. ipython:: python

            sdr.ebn0_to_esn0(5, 2, rate=2/3)

        Convert from $E_b/N_0 = 10$ dB to $E_s/N_0$ for a 16-QAM signal with $r = 1$.

        .. ipython:: python

            sdr.ebn0_to_esn0(10, 4, rate=1)

    Group:
        conversions-snrs
    """
    ebn0 = verify_arraylike(ebn0)  # Energy per information bit
    bps = verify_arraylike(bps, int=True, positive=True)
    rate = verify_arraylike(rate, float=True, inclusive_min=0, inclusive_max=1)

    ecn0 = ebn0 + db(rate)  # Energy per coded bit
    esn0 = ecn0 + db(bps)  # Energy per symbol

    return convert_output(esn0)


@export
def ebn0_to_snr(
    ebn0: npt.ArrayLike,
    bps: npt.ArrayLike,
    rate: npt.ArrayLike = 1.0,
    sps: npt.ArrayLike = 1,
) -> npt.NDArray[np.float64]:
    r"""
    Converts from $E_b/N_0$ to $S/N$.

    $$
    \frac{S}{N} = \frac{E_b}{N_0} \frac{k}{n} \log_2 M \frac{f_{sym}}{f_s}
    $$

    Arguments:
        ebn0: Bit energy $E_b$ to noise PSD $N_0$ ratio in dB.
        bps: Coded bits per symbol $\log_2 M$, where $M$ is the modulation order.
        rate: Code rate $r = k/n$, where $k$ is the number of information bits and $n$ is the
            number of coded bits.
        sps: Samples per symbol $f_s / f_{sym}$.

    Returns:
        The signal-to-noise ratio $S/N$ in dB.

    Examples:
        Convert from $E_b/N_0 = 5$ dB to $S/N$ for a 4-QAM signal with $r = 2/3$ and 1 sample per symbol.

        .. ipython:: python

            sdr.ebn0_to_snr(5, 2, rate=2/3, sps=1)

        Convert from $E_b/N_0 = 10$ dB to $S/N$ for a 16-QAM signal with $r = 1$ and 4 samples per symbol.

        .. ipython:: python

            sdr.ebn0_to_snr(10, 4, rate=1, sps=4)

    Group:
        conversions-snrs
    """
    ebn0 = verify_arraylike(ebn0)  # Energy per information bit
    bps = verify_arraylike(bps, int=True, positive=True)
    rate = verify_arraylike(rate, float=True, inclusive_min=0, inclusive_max=1)
    sps = verify_arraylike(sps, int=True, positive=True)

    esn0 = ebn0_to_esn0(ebn0, bps, rate=rate)  # SNR per symbol
    snr = esn0 - db(sps)  # SNR per sample

    return convert_output(snr)


##############################################################################
# From Es/N0
##############################################################################


@export
def esn0_to_ebn0(
    esn0: npt.ArrayLike,
    bps: npt.ArrayLike,
    rate: npt.ArrayLike = 1.0,
) -> npt.NDArray[np.float64]:
    r"""
    Converts from $E_s/N_0$ to $E_b/N_0$.

    $$
    \frac{E_b}{N_0} = \frac{E_s}{N_0} \frac{n}{k} \frac{1}{\log_2 M}
    $$

    Arguments:
        esn0: Symbol energy $E_s$ to noise PSD $N_0$ ratio in dB.
        bps: Coded bits per symbol $\log_2 M$, where $M$ is the modulation order.
        rate: Code rate $r = k/n$, where $k$ is the number of information bits and $n$ is the
            number of coded bits.

    Returns:
        The bit energy $E_b$ to noise PSD $N_0$ ratio in dB.

    Examples:
        Convert from $E_s/N_0 = 5$ dB to $E_b/N_0$ for a 4-QAM signal with $r = 2/3$.

        .. ipython:: python

            sdr.esn0_to_ebn0(5, 2, rate=2/3)

        Convert from $E_s/N_0 = 10$ dB to $E_b/N_0$ for a 16-QAM signal with $r = 1$.

        .. ipython:: python

            sdr.esn0_to_ebn0(10, 4, rate=1)

    Group:
        conversions-snrs
    """
    esn0 = verify_arraylike(esn0)  # Energy per symbol
    bps = verify_arraylike(bps, int=True, positive=True)
    rate = verify_arraylike(rate, float=True, inclusive_min=0, inclusive_max=1)

    ecn0 = esn0 - db(bps)  # Energy per coded bit
    ebn0 = ecn0 - db(rate)  # Energy per information bit

    return convert_output(ebn0)


@export
def esn0_to_snr(
    esn0: npt.ArrayLike,
    sps: npt.ArrayLike = 1,
) -> npt.NDArray[np.float64]:
    r"""
    Converts from $E_s/N_0$ to $S/N$.

    $$
    \frac{S}{N} = \frac{E_s}{N_0} \frac{f_{sym}}{f_s}
    $$

    Arguments:
        esn0: Symbol energy $E_s$ to noise PSD $N_0$ ratio in dB.
        sps: Samples per symbol $f_s / f_{sym}$.

    Returns:
        The signal-to-noise ratio $S/N$ in dB.

    Examples:
        Convert from $E_s/N_0 = 5$ dB to $S/N$ with 1 sample per symbol. In discrete-time systems,
        when there is 1 sample per symbol, $S/N$ is equivalent to $E_s/N_0$.

        .. ipython:: python

            sdr.esn0_to_snr(5, sps=1)

        Convert from $E_s/N_0 = 10$ dB to $S/N$ with 4 samples per symbol.

        .. ipython:: python

            sdr.esn0_to_snr(10, sps=4)

    Group:
        conversions-snrs
    """
    esn0 = verify_arraylike(esn0)  # Energy per symbol
    sps = verify_arraylike(sps, int=True, positive=True)

    snr = esn0 - db(sps)  # SNR per sample

    return convert_output(snr)


##############################################################################
# From SNR
##############################################################################


@export
def snr_to_ebn0(
    snr: npt.ArrayLike,
    bps: npt.ArrayLike,
    rate: npt.ArrayLike = 1.0,
    sps: npt.ArrayLike = 1,
) -> npt.NDArray[np.float64]:
    r"""
    Converts from $S/N$ to $E_b/N_0$.

    $$
    \frac{E_b}{N_0} = \frac{S}{N} \frac{f_s}{f_{sym}} \frac{n}{k} \frac{1}{\log_2 M}
    $$

    Arguments:
        snr: Signal-to-noise ratio $S/N$ in dB.
        bps: Coded bits per symbol $\log_2 M$, where $M$ is the modulation order.
        rate: Code rate $r = k/n$, where $k$ is the number of information bits and $n$ is the
            number of coded bits.
        sps: Samples per symbol $f_s / f_{sym}$.

    Returns:
        The bit energy $E_b$ to noise PSD $N_0$ ratio in dB.

    Examples:
        Convert from $S/N = 5$ dB to $E_b/N_0$ for a 4-QAM signal with $r = 2/3$ and 1 sample per symbol.

        .. ipython:: python

            sdr.snr_to_ebn0(5, 2, rate=2/3, sps=1)

        Convert from $S/N = 10$ dB to $E_b/N_0$ for a 16-QAM signal with $r = 1$ and 4 samples per symbol.

        .. ipython:: python

            sdr.snr_to_ebn0(10, 4, rate=1, sps=4)

    Group:
        conversions-snrs
    """
    snr = verify_arraylike(snr)  # SNR per sample
    bps = verify_arraylike(bps, int=True, positive=True)
    rate = verify_arraylike(rate, float=True, inclusive_min=0, inclusive_max=1)
    sps = verify_arraylike(sps, int=True, positive=True)

    esn0 = snr_to_esn0(snr, sps=sps)  # Energy per symbol
    ebn0 = esn0_to_ebn0(esn0, bps, rate=rate)  # Energy per information bit

    return convert_output(ebn0)


@export
def snr_to_esn0(
    snr: npt.ArrayLike,
    sps: npt.ArrayLike = 1,
) -> npt.NDArray[np.float64]:
    r"""
    Converts from $S/N$ to $E_s/N_0$.

    $$
    \frac{E_s}{N_0} = \frac{S}{N} \frac{f_s}{f_{sym}}
    $$

    Arguments:
        snr: Signal-to-noise ratio $S/N$ in dB.
        sps: Samples per symbol $f_s / f_{sym}$.

    Returns:
        The symbol energy $E_s$ to noise PSD $N_0$ ratio in dB.

    Examples:
        Convert from $S/N = 5$ dB to $E_s/N_0$ with 1 sample per symbol. In discrete-time systems,
        when there is 1 sample per symbol, $S/N$ is equivalent to $E_s/N_0$.

        .. ipython:: python

            sdr.snr_to_esn0(5, sps=1)

        Convert from $S/N = 10$ dB to $E_s/N_0$ with 4 samples per symbol.

        .. ipython:: python

            sdr.snr_to_esn0(10, sps=4)

    Group:
        conversions-snrs
    """
    snr = verify_arraylike(snr)  # SNR per sample
    sps = verify_arraylike(sps, int=True, positive=True)

    esn0 = snr + db(sps)  # Energy per symbol

    return convert_output(esn0)



import galois
import numpy as np

import sdr


def test_ieee802_11():
    c = galois.Poly.Degrees([7, 3, 0])
    scrambler = sdr.AdditiveScrambler(c)
    x = np.zeros(127, dtype=int)  # If input is all zeros, output is the scrambling sequence
    y = scrambler.scramble(x)
    seq_truth = np.array([0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1])  # fmt: skip
    assert np.array_equal(y, seq_truth)


def test_scramble_descramble():
    c = galois.Poly.Degrees([7, 3, 0])
    scrambler = sdr.AdditiveScrambler(c)
    rng = np.random.default_rng()
    x = rng.integers(0, 2, 1000)
    y = scrambler.scramble(x)
    x_recovered = scrambler.descramble(y)
    assert np.array_equal(x, x_recovered)



import numpy as np

import sdr


def test_da_error():
    ped = sdr.PED()
    modem = sdr.PSK(4)
    error, da_error = ped.data_aided_error(modem)
    da_error_truth = error
    assert np.allclose(da_error, da_error_truth)



"""
A module containing utility functions for plotting.
"""

from __future__ import annotations

from typing import overload

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from .._helper import (
    export,
    verify_arraylike,
    verify_isinstance,
    verify_positional_args,
)
from ._rc_params import RC_PARAMS


@overload
def stem(
    x: npt.NDArray,  # TODO: Change to npt.ArrayLike once Sphinx has better overload support
    *,
    color: str | None = None,
    ax: plt.Axes | None = None,
    **kwargs,
): ...


@overload
def stem(
    x: npt.NDArray,  # TODO: Change to npt.ArrayLike once Sphinx has better overload support
    y: npt.NDArray,  # TODO: Change to npt.ArrayLike once Sphinx has better overload support
    *,
    color: str | None = None,
    ax: plt.Axes | None = None,
    **kwargs,
): ...


@export
def stem(  # noqa: D417
    *args,
    color: str | None = None,
    ax: plt.Axes | None = None,
    **kwargs,
):
    r"""
    Wraps :func:`matplotlib.pyplot.stem` to style the plot more like MATLAB.

    Arguments:
        x: The x-coordinates of the stem plot.
        y: The y-coordinates of the stem plot.
        color: The color of the stem line and marker. If `None`, the next color in the current color cycle is used.
        ax: The axis to plot on. If `None`, the current axis is used.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.stem`.

    Notes:
        This function is a wrapper around :func:`matplotlib.pyplot.stem` that styles the plot more like MATLAB's
        `stem()`. It removes the base line and sets the stem line color and marker color to the next color in the
        current color cycle. It also sets the marker face color to none.

    Examples:
        Compare the :func:`matplotlib.pyplot.stem` plot to the styled :func:`sdr.plot.stem` plot.

        .. ipython:: python

            rrc = sdr.raised_cosine(0.1, 8, 10)

            @savefig sdr_plot_stem_1.png
            plt.figure(); \
            plt.stem(rrc); \
            plt.title("matplotlib.pyplot.stem()");

            @savefig sdr_plot_stem_2.png
            plt.figure(); \
            sdr.plot.stem(rrc); \
            plt.title("sdr.plot.stem()");

        The standard plot is even more unreadable when multiple stem plots are on the same axes.

        .. ipython:: python

            gaussian = sdr.gaussian(0.1, 8, 10)

            @savefig sdr_plot_stem_3.png
            plt.figure(); \
            plt.stem(rrc); \
            plt.stem(gaussian); \
            plt.title("matplotlib.pyplot.stem()");

            @savefig sdr_plot_stem_4.png
            plt.figure(); \
            sdr.plot.stem(rrc); \
            sdr.plot.stem(gaussian); \
            plt.title("sdr.plot.stem()");

    Group:
        plot-utility
    """
    verify_positional_args(args, 2)
    if len(args) == 1:
        y = verify_arraylike(args[0], complex=True, ndim=1)
        x = np.arange(y.size)
    elif len(args) == 2:
        x = verify_arraylike(args[0], float=True, ndim=1)
        y = verify_arraylike(args[1], complex=True, ndim=1)
    verify_isinstance(color, str, optional=True)
    verify_isinstance(ax, plt.Axes, optional=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        # Make a dummy plot to get the next color in the current color cycle
        if color is None:
            (line,) = ax.plot([], [])
            color = line.get_color()

        # Plot matplotlib's standard stem plot
        markerline, stemlines, baseline = ax.stem(x, y, **kwargs)

        # Style the stem plot more like MATLAB
        plt.setp(markerline, "markerfacecolor", "none", "markeredgecolor", color)
        plt.setp(stemlines, "color", color, "linewidth", 1)
        plt.setp(baseline, "color", "none", "linewidth", 1)



"""
A module containing functions related to non-coherent integration.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.interpolate
import scipy.optimize
from typing_extensions import Literal

from .._conversion import db
from .._helper import convert_output, export, verify_arraylike, verify_bool, verify_literal
from ._theory import p_d


@export
def non_coherent_gain(
    n_nc: npt.ArrayLike,
    snr: npt.ArrayLike,
    p_fa: npt.ArrayLike = 1e-6,
    detector: Literal["linear", "square-law"] = "square-law",
    complex: bool = True,
    snr_ref: Literal["input", "output"] = "input",
    extrapolate: bool = True,
) -> npt.NDArray[np.float64]:
    r"""
    Computes the SNR improvement by non-coherently integrating $N_{nc}$ samples.

    Arguments:
        n_nc: The number of samples $N_{nc}$ to non-coherently integrate.
        snr: The reference SNR in dB.
        p_fa: The desired probability of false alarm $P_{fa}$. This is used to compute the necessary thresholds before
            and after integration. The non-coherent gain is slightly affected by the $P_{fa}$.
        detector: The detector type.

            - `"coherent"`: A coherent detector, $$T(x) = \mathrm{Re}\left\{\sum_{i=0}^{N_c-1} x[n-i]\right\} .$$
            - `"linear"`: A linear detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right| .$$
            - `"square-law"`: A square-law detector, $$T(x) = \sum_{j=0}^{N_{nc}-1}\left|\sum_{i=0}^{N_c-1} x[n-i-jN_c]\right|^2 .$$

        complex: Indicates whether the input signal is real or complex. This affects how the SNR is converted
            to noise variance.
        snr_ref: The SNR reference.

            - `"input"`: The SNR is referenced at the input of the non-coherent integrator $\text{SNR}_{\text{in},N_{nc}}$.
            - `"output"`: The SNR is referenced at the output of the non-coherent integrator $\text{SNR}_{\text{in},1}$.

        extrapolate: Indicates whether to extrapolate $G_{nc}$ using smaller values of $N_{nc}$. This is only done when
            the non-coherent gain cannot be explicitly solved for due to lack of floating-point precision.
            If `False`, the function will return `np.nan` for any $N_{nc}$ that cannot be solved for.

    Returns:
        The non-coherent gain $G_{nc}$ in dB.

    See Also:
        sdr.peebles

    Notes:
        Let the input SNR required to achieve the specified detection performance with no non-coherent integration
        be $\text{SNR}_{\text{in},1}$. Let the input SNR required to achieve the same detection performance with
        $N_{nc}$ non-coherent integrations be $\text{SNR}_{\text{in},N_{nc}}$. The non-coherent gain is defined as
        their difference. It is the SNR reduction provided by the use of non-coherent integration.

        $$G_{nc} = \text{SNR}_{\text{in},1} - \text{SNR}_{\text{in},N_{nc}}$$

    Examples:
        See the :ref:`non-coherent-integration` example.

        Compute the non-coherent gain for various integration lengths at 10-dB SNR.

        .. ipython:: python

            sdr.non_coherent_gain(1, 10)
            sdr.non_coherent_gain(2, 10)
            sdr.non_coherent_gain(10, 10)
            sdr.non_coherent_gain(20, 10)

        Plot the non-coherent gain parameterized by input SNR.

        .. ipython:: python

            plt.figure(); \
            n = np.logspace(0, 3, 51).astype(int); \
            plt.semilogx(n, sdr.coherent_gain(n), color="k");
            for snr in np.arange(-30, 40, 10):
                plt.semilogx(n, sdr.non_coherent_gain(n, snr, snr_ref="input"), label=f"{snr} dB")
            @savefig sdr_non_coherent_gain_1.png
            plt.legend(title="Input SNR", loc="upper left"); \
            plt.xlabel("Number of samples, $N_{nc}$"); \
            plt.ylabel("Non-coherent gain, $G_{nc}$"); \
            plt.title("Non-coherent gain for various input SNRs");

        Compare the non-coherent gain for linear and square-law detectors for various input SNRs. Notice that the
        square-law detector performs better at low SNRs and the linear detector performs better at high SNRs.

        .. ipython:: python

            @savefig sdr_non_coherent_gain_2.png
            plt.figure(); \
            snr = np.linspace(-20, 10, 21); \
            plt.plot(snr, sdr.non_coherent_gain(10, snr, p_fa=1e-6, detector="linear"), label="Linear"); \
            plt.plot(snr, sdr.non_coherent_gain(10, snr, p_fa=1e-6, detector="square-law"), label="Square-law"); \
            plt.legend(title="Detector", loc="upper left"); \
            plt.xlabel("Input signal-to-noise ratio (dB)"); \
            plt.ylabel("Non-coherent gain, $G_{nc}$"); \
            plt.title("Non-coherent gain for $N_{nc} = 10$");

        Examine the non-coherent gain across input SNR and false alarm rate for non-coherently integrating 10 samples.
        Notice that the non-coherent gain is affected by both. The coherent integration gain, however, is
        a constant 10 dB across both.

        .. ipython:: python

            plt.figure(); \
            snr = np.linspace(-40, 12, 101); \
            n_nc = 10;
            for p_fa in [1e-14, 1e-12, 1e-10, 1e-8, 1e-6, 1e-4, 1e-2]:
                g_nc = sdr.non_coherent_gain(n_nc, snr, p_fa)
                plt.plot(snr, g_nc, label=f"{p_fa:1.0e}")
            @savefig sdr_non_coherent_gain_3.png
            plt.legend(title="$P_{fa}$"); \
            plt.ylim(0, 10); \
            plt.xlabel("Input signal-to-noise ratio (dB)"); \
            plt.ylabel("Non-coherent gain (dB)"); \
            plt.title("Non-coherent gain for $N_{nc} = 10$");

    Group:
        detection-non-coherent-integration
    """
    n_nc = verify_arraylike(n_nc, dtype=object, positive=True)  # TODO: Downstream functions want a Python int?
    snr = verify_arraylike(snr, float=True)
    p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)
    verify_literal(detector, ["linear", "square-law"])
    verify_bool(complex)
    verify_literal(snr_ref, ["input", "output"])
    verify_bool(extrapolate)

    if snr_ref == "input":
        g_nc = _non_coherent_gain_in(n_nc, snr, p_fa, detector, complex)
    else:
        g_nc = _non_coherent_gain_out(n_nc, snr, p_fa, detector, complex)

    if extrapolate:
        g_nc = _extrapolate_non_coherent_gain(n_nc, snr, p_fa, g_nc, snr_ref, detector, complex)

    return convert_output(g_nc)


@np.vectorize
def _non_coherent_gain_in(
    n_nc: int,
    snr: float,
    p_fa: float,
    detector: Literal["linear", "square-law"],
    complex: bool,
) -> float:
    """
    Solves for the non-coherent gain when the SNR is referenced at the input of the non-coherent integrator.
    """
    if n_nc == 1:
        return 0.0

    # Compute the theoretical p_d with the following input SNR and n_nc non-coherent combinations
    p_d_in = p_d(snr, p_fa, detector=detector, complex=complex, n_c=1, n_nc=n_nc)
    if p_d_in == 1:
        return np.nan

    def root_eq(snr_out: float):
        p_d_out = p_d(snr_out, p_fa, detector=detector, complex=complex, n_c=1, n_nc=1)
        return db(p_d_out) - db(p_d_in)  # Use logarithms for numerical stability

    # Determine the output SNR that, without non-coherent integration, yields the same probability of detection.
    snr_out = scipy.optimize.brentq(root_eq, snr, snr + db(n_nc))  # Bounds are 0 to coherent gain
    g_nc = snr_out - snr

    return g_nc


@np.vectorize
def _non_coherent_gain_out(
    n_nc: int,
    snr: float,
    p_fa: float,
    detector: Literal["linear", "square-law"],
    complex: bool,
) -> float:
    """
    Solves for the non-coherent gain when the SNR is referenced at the output of the non-coherent integrator.
    """
    if n_nc == 1:
        return 0.0

    # Compute the theoretical p_d with the following output SNR and no non-coherent combinations
    p_d_out = p_d(snr, p_fa, detector=detector, complex=complex, n_c=1, n_nc=1)
    if p_d_out == 1:
        return np.nan

    def root_eq(snr_in: float):
        p_d_in = p_d(snr_in, p_fa, detector=detector, complex=complex, n_c=1, n_nc=n_nc)
        return db(p_d_out) - db(p_d_in)  # Use logarithms for numerical stability

    # Determine the input SNR that, after non-coherent integration, yields the same probability of detection.
    snr_in = scipy.optimize.brentq(root_eq, snr - db(n_nc), snr)  # Bounds are 0 to coherent gain
    g_nc = snr - snr_in

    return g_nc


def _extrapolate_non_coherent_gain(
    n_nc: npt.NDArray[np.float64],
    snr: npt.NDArray[np.float64],
    p_fa: npt.NDArray[np.float64],
    g_nc: npt.NDArray[np.float64],
    snr_ref: Literal["input", "output"],
    detector: Literal["linear", "square-law"],
    complex: bool,
) -> npt.NDArray[np.float64]:
    """
    Computes the non-coherent gain for smaller n_nc values and extrapolates.
    """
    # Broadcast arrays to the same shape and then flatten
    n_nc, snr, p_fa, g_nc = np.broadcast_arrays(n_nc, snr, p_fa, g_nc)
    shape = g_nc.shape
    n_nc = n_nc.ravel()
    snr = snr.ravel()
    p_fa = p_fa.ravel()
    g_nc = g_nc.ravel()

    # Loop through the NaNs and interpolate
    interpolators = {}
    for i in range(g_nc.size):
        if np.isnan(g_nc[i]) and n_nc[i] > 1:  # Only interpolate where the gain is NaN and n_nc is greater than 1
            key = (snr[i], p_fa[i])
            if key not in interpolators:
                n_nc_array = np.logspace(np.log10(max(1, 0.5 * n_nc[i])), np.log10(n_nc[i]), 7).astype(int)
                n_nc_array = np.unique(n_nc_array)  # Remove unhelpful duplicates
                g_nc_array = non_coherent_gain(n_nc_array, snr[i], p_fa[i], detector, complex, snr_ref, False)

                idxs = ~np.isnan(g_nc_array)
                n_nc_array = n_nc_array[idxs]
                g_nc_array = g_nc_array[idxs]

                if n_nc_array.size < 2:
                    # There aren't enough data points to interpolate
                    continue

                interpolators[key] = scipy.interpolate.interp1d(
                    np.log10(n_nc_array), g_nc_array, kind="linear", fill_value="extrapolate"
                )

            g_nc[i] = interpolators[key](np.log10(n_nc[i]))

    g_nc = g_nc.reshape(shape)

    return g_nc



import numpy as np
import pytest

import sdr


def test_scalar():
    assert sdr.max_integration_time(3, 235) == pytest.approx(0.0018818867640235891)


def test_cgl_vector():
    t = sdr.max_integration_time([1, 2, 3, 4, 5], 100)
    t_truth = np.array([0.002615, 0.00365466, 0.00442243, 0.00504438, 0.00556994])
    assert isinstance(t, np.ndarray)
    assert np.allclose(t, t_truth)


def test_freq_vector():
    t = sdr.max_integration_time(3, [100, 200, 300, 400, 500])
    t_truth = np.array([0.00442243, 0.00221122, 0.00147414, 0.00110561, 0.00088449])
    assert isinstance(t, np.ndarray)
    assert np.allclose(t, t_truth)



"""
A module containing :obj:`sdr`'s default matplotlib rcParams.
"""

from __future__ import annotations

import matplotlib.pyplot as plt

from .._helper import export

RC_PARAMS = {
    "lines.linewidth": 1,
    "axes.grid": True,
    "axes.grid.which": "both",
    # "axes.xmargin": 0,
    # "axes.ymargin": 0,
    # "axes.zmargin": 0,
    "figure.constrained_layout.use": True,
    "figure.constrained_layout.h_pad": 0.1,
    "figure.constrained_layout.w_pad": 0.1,
    "figure.figsize": (8, 4),
    "figure.max_open_warning": 0,
    "figure.titleweight": "bold",
    "grid.alpha": 0.8,
    "grid.linestyle": "--",
}


@export
def use_style():
    """
    Applies :obj:`sdr`'s default :obj:`matplotlib` rcParams.

    These style settings may be reverted with :func:`matplotlib.pyplot.rcdefaults`.

    Examples:
        The following rcParams are applied.

        .. ipython:: python

            sdr.plot._rc_params.RC_PARAMS

    Group:
        plot-utility
    """
    plt.rcParams.update(RC_PARAMS)



"""
A module containing filter-related plotting functions.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import scipy.signal
from typing_extensions import Literal

from .._conversion import db
from .._filter import FIR, IIR
from .._helper import export, verify_arraylike, verify_bool, verify_isinstance, verify_literal, verify_scalar
from ._helper import integer_x_axis, min_ylim, standard_plot, verify_sample_rate
from ._rc_params import RC_PARAMS
from ._units import freq_units, time_units


def _convert_to_taps(
    filter: FIR | IIR | npt.ArrayLike | tuple[npt.ArrayLike, npt.ArrayLike],
) -> tuple[npt.NDArray, npt.NDArray]:
    if isinstance(filter, FIR):
        b = filter.taps
        a = np.array([1])
    elif isinstance(filter, IIR):
        b = filter.b_taps
        a = filter.a_taps
    elif isinstance(filter, tuple):
        b = verify_arraylike(filter[0], complex=True, ndim=1)
        a = verify_arraylike(filter[1], complex=True, ndim=1)
    else:
        b = verify_arraylike(filter, complex=True, ndim=1)
        a = np.array([1])
    return b, a


@export
def impulse_response(
    filter: FIR | IIR | npt.ArrayLike | tuple[npt.ArrayLike, npt.ArrayLike],
    N: int | None = None,
    offset: float = 0.0,
    ax: plt.Axes | None = None,
    type: Literal["plot", "stem"] = "stem",
    **kwargs,
):
    r"""
    Plots the impulse response $h[n]$ of a filter.

    The impulse response $h[n]$ is the filter output when the input is an impulse $\delta[n]$.

    Arguments:
        filter: The filter definition.

            - :class:`sdr.FIR`, :class:`sdr.IIR`: The filter object.
            - `npt.ArrayLike`: The feedforward coefficients $b_i$.
            - `tuple[npt.ArrayLike, npt.ArrayLike]`: The feedforward coefficients $b_i$ and
              feedback coefficients $a_j$.

        N: The number of samples $N$ to plot. If `None`, the length of `b` is used for FIR filters and
            100 for IIR filters.
        offset: The x-axis offset to apply to the first sample. Can be useful for comparing the impulse
            response of filters with different lengths.
        ax: The axis to plot on. If `None`, the current axis is used.
        type: The type of plot to use.
        kwargs: Additional keyword arguments to pass to the plotting function.

    Examples:
        See the :ref:`fir-filters` example.

        .. ipython:: python

            h_srrc = sdr.root_raised_cosine(0.5, 10, 10)

            @savefig sdr_plot_impulse_response_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h_srrc)

        See the :ref:`iir-filters` example.

        .. ipython:: python

            zero = 0.6; \
            pole = 0.8 * np.exp(1j * np.pi / 8); \
            iir = sdr.IIR.ZerosPoles([zero], [pole, pole.conj()])

            @savefig sdr_plot_impulse_response_2.png
            plt.figure(); \
            sdr.plot.impulse_response(iir, N=30)

    Group:
        plot-filter
    """
    b, a = _convert_to_taps(filter)
    verify_scalar(N, optional=True, int=True, positive=True)
    verify_scalar(offset, float=True)
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(type, ["plot", "stem"])

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if N is None:
            if a.size == 1 and a[0] == 1:
                N = b.size  # FIR filter
            else:
                N = 100  # IIR filter

        # Delta impulse function
        d = np.zeros(N, dtype=float)
        d[0] = 1

        # Filter the impulse
        zi = scipy.signal.lfiltic(b, a, y=[], x=[])
        h, zi = scipy.signal.lfilter(b, a, d, zi=zi)
        t = np.arange(h.size) + offset

        integer_x_axis(ax)
        standard_plot(t, h, ax=ax, type=type, y_axis="complex", **kwargs)
        ax.set_xlabel("Sample, $n$")
        ax.set_ylabel("Amplitude")
        ax.set_title("Impulse response, $h[n]$")


@export
def step_response(
    filter: FIR | IIR | npt.ArrayLike | tuple[npt.ArrayLike, npt.ArrayLike],
    N: int | None = None,
    ax: plt.Axes | None = None,
    type: Literal["plot", "stem"] = "stem",
    **kwargs,
):
    r"""
    Plots the step response $s[n]$ of a filter.

    The step response $s[n]$ is the filter output when the input is a unit step $u[n]$.

    Arguments:
        filter: The filter definition.

            - :class:`sdr.FIR`, :class:`sdr.IIR`: The filter object.
            - `npt.ArrayLike`: The feedforward coefficients $b_i$.
            - `tuple[npt.ArrayLike, npt.ArrayLike]`: The feedforward coefficients $b_i$ and
              feedback coefficients $a_j$.

        N: The number of samples $N$ to plot. If `None`, the length of `b` is used for FIR filters and
            100 for IIR filters.
        ax: The axis to plot on. If `None`, the current axis is used.
        type: The type of plot to use.
        kwargs: Additional keyword arguments to pass to the plotting function.

    Examples:
        See the :ref:`fir-filters` example.

        .. ipython:: python

            h_srrc = sdr.root_raised_cosine(0.5, 10, 10)

            @savefig sdr_plot_step_response_1.png
            plt.figure(); \
            sdr.plot.step_response(h_srrc)

        See the :ref:`iir-filters` example.

        .. ipython:: python

            zero = 0.6; \
            pole = 0.8 * np.exp(1j * np.pi / 8); \
            iir = sdr.IIR.ZerosPoles([zero], [pole, pole.conj()])

            @savefig sdr_plot_step_response_2.png
            plt.figure(); \
            sdr.plot.step_response(iir, N=30)

    Group:
        plot-filter
    """
    b, a = _convert_to_taps(filter)
    verify_scalar(N, optional=True, int=True, positive=True)
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(type, ["plot", "stem"])

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if N is None:
            if a.size == 1 and a[0] == 1:
                N = b.size  # FIR filter
            else:
                N = 100  # IIR filter

        # Unit step function
        u = np.ones(N, dtype=float)

        # Filter the impulse
        zi = scipy.signal.lfiltic(b, a, y=[], x=[])
        s, zi = scipy.signal.lfilter(b, a, u, zi=zi)
        t = np.arange(s.size)

        integer_x_axis(ax)
        standard_plot(t, s, ax=ax, type=type, y_axis="complex", **kwargs)
        ax.set_xlabel("Sample, $n$")
        ax.set_ylabel("Amplitude")
        ax.set_title("Step response, $s[n]$")


@export
def zeros_poles(
    filter: FIR | IIR | npt.ArrayLike | tuple[npt.ArrayLike, npt.ArrayLike],
    ax: plt.Axes | None = None,
    **kwargs,
):
    r"""
    Plots the zeros and poles of the filter.

    Arguments:
        filter: The filter definition.

            - :class:`sdr.FIR`, :class:`sdr.IIR`: The filter object.
            - `npt.ArrayLike`: The feedforward coefficients $b_i$.
            - `tuple[npt.ArrayLike, npt.ArrayLike]`: The feedforward coefficients $b_i$ and
              feedback coefficients $a_j$.

        ax: The axis to plot on. If `None`, the current axis is used.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

    Examples:
        See the :ref:`fir-filters` example.

        .. ipython:: python

            h_srrc = sdr.root_raised_cosine(0.5, 10, 10)

            @savefig sdr_plot_zeros_poles_1.png
            plt.figure(); \
            sdr.plot.zeros_poles(h_srrc)

        See the :ref:`iir-filters` example.

        .. ipython:: python

            zero = 0.6; \
            pole = 0.8 * np.exp(1j * np.pi / 8); \
            iir = sdr.IIR.ZerosPoles([zero], [pole, pole.conj()])

            @savefig sdr_plot_zeros_poles_2.png
            plt.figure(); \
            sdr.plot.zeros_poles(iir)

    Group:
        plot-filter
    """
    b, a = _convert_to_taps(filter)
    verify_isinstance(ax, plt.Axes, optional=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        z, p, _ = scipy.signal.tf2zpk(b, a)
        unit_circle = np.exp(1j * np.linspace(0, 2 * np.pi, 100))

        label = kwargs.pop("label", None)
        if label is None:
            z_label = "Zeros"
            p_label = "Poles"
        else:
            z_label = label + " (zeros)"
            p_label = label + " (poles)"

        ax.plot(unit_circle.real, unit_circle.imag, color="k", linestyle="--", label="Unit circle")
        ax.scatter(z.real, z.imag, marker="o", label=z_label)
        ax.scatter(p.real, p.imag, marker="x", label=p_label)
        ax.axis("equal")
        ax.legend()
        ax.set_xlabel("Real")
        ax.set_ylabel("Imaginary")
        ax.set_title("Zeros and poles of $H(z)$")


@export
def magnitude_response(
    filter: FIR | IIR | npt.ArrayLike | tuple[npt.ArrayLike, npt.ArrayLike],
    sample_rate: float | None = None,
    N: int = 1024,
    ax: plt.Axes | None = None,
    x_axis: Literal["auto", "one-sided", "two-sided", "log"] = "auto",
    y_axis: Literal["linear", "log"] = "log",
    decades: int = 4,
    **kwargs,
):
    r"""
    Plots the magnitude response $|H(\omega)|^2$ of the filter.

    Arguments:
        filter: The filter definition.

            - :class:`sdr.FIR`, :class:`sdr.IIR`: The filter object.
            - `npt.ArrayLike`: The feedforward coefficients $b_i$.
            - `tuple[npt.ArrayLike, npt.ArrayLike]`: The feedforward coefficients $b_i$ and
              feedback coefficients $a_j$.

        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Normalized frequency".
        N: The number of samples $N$ in the frequency response.
        ax: The axis to plot on. If `None`, the current axis is used.
        x_axis: The x-axis scaling. Options are to display a one-sided spectrum, a two-sided spectrum, or
            one-sided spectrum with a logarithmic frequency axis. The default is `"auto"` which selects `"one-sided"`
            for real-valued filters and `"two-sided"` for complex-valued filters.
        y_axis: The y-axis scaling. Options are to display a linear or logarithmic magnitude response.
        decades: The number of decades to plot when `x_axis="log"`.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

    Examples:
        See the :ref:`fir-filters` example.

        .. ipython:: python

            h_srrc = sdr.root_raised_cosine(0.5, 10, 10)

            @savefig sdr_plot_magnitude_response_1.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_srrc)

        See the :ref:`iir-filters` example.

        .. ipython:: python

            zero = 0.6; \
            pole = 0.8 * np.exp(1j * np.pi / 8); \
            iir = sdr.IIR.ZerosPoles([zero], [pole, pole.conj()])

            @savefig sdr_plot_magnitude_response_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(iir)

        .. ipython:: python

            @savefig sdr_plot_magnitude_response_3.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_srrc, x_axis="two-sided")

        .. ipython:: python

            @savefig sdr_plot_magnitude_response_4.png
            plt.figure(); \
            sdr.plot.magnitude_response(iir, x_axis="log", decades=3)

    Group:
        plot-filter
    """
    b, a = _convert_to_taps(filter)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate)
    verify_scalar(N, optional=True, int=True, positive=True)
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(x_axis, ["auto", "one-sided", "two-sided", "log"])
    verify_literal(y_axis, ["linear", "log"])
    verify_scalar(decades, int=True, positive=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if x_axis == "auto":
            x_axis = "one-sided" if np.isrealobj(b) and np.isrealobj(a) else "two-sided"

        with np.errstate(divide="ignore", invalid="ignore"):
            if x_axis == "log":
                f = np.logspace(np.log10(sample_rate / 2 / 10**decades), np.log10(sample_rate / 2), N)
                f, H = scipy.signal.freqz(b, a, worN=f, whole=False, fs=sample_rate)
            else:
                f, H = scipy.signal.freqz(b, a, worN=N, whole=x_axis == "two-sided", fs=sample_rate)

        if x_axis == "two-sided":
            f -= sample_rate / 2
            H = np.fft.fftshift(H)

        if sample_rate_provided:
            units, scalar = freq_units(f)
            f *= scalar

        if y_axis == "log":
            H = db(np.abs(H) ** 2)
        else:
            H = np.abs(H) ** 2

        if x_axis == "log":
            ax.semilogx(f, H, **kwargs)
        else:
            ax.plot(f, H, **kwargs)

        ax.grid(True, which="both")
        if "label" in kwargs:
            ax.legend()

        if sample_rate_provided:
            ax.set_xlabel(f"Frequency ({units}), $f$")
        else:
            ax.set_xlabel("Normalized frequency, $f / f_s$")

        if y_axis == "log":
            ax.set_ylabel(r"Power (dB), $|H(\omega)|^2$")
        else:
            ax.set_ylabel(r"Power, $|H(\omega)|^2$")

        ax.set_title(r"Magnitude response, $|H(\omega)|^2$")


@export
def phase_response(
    filter: FIR | IIR | npt.ArrayLike | tuple[npt.ArrayLike, npt.ArrayLike],
    sample_rate: float | None = None,
    N: int = 1024,
    unwrap: bool = True,
    ax: plt.Axes | None = None,
    x_axis: Literal["auto", "one-sided", "two-sided", "log"] = "auto",
    decades: int = 4,
    **kwargs,
):
    r"""
    Plots the phase response $\angle H(\omega)$ of the filter.

    Arguments:
        filter: The filter definition.

            - :class:`sdr.FIR`, :class:`sdr.IIR`: The filter object.
            - `npt.ArrayLike`: The feedforward coefficients $b_i$.
            - `tuple[npt.ArrayLike, npt.ArrayLike]`: The feedforward coefficients $b_i$ and
              feedback coefficients $a_j$.

        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Normalized frequency".
        N: The number of samples $N$ in the phase response.
        unwrap: Indicates whether to unwrap the phase response.
        ax: The axis to plot on. If `None`, the current axis is used.
        x_axis: The x-axis scaling. Options are to display a one-sided spectrum, a two-sided spectrum, or
            one-sided spectrum with a logarithmic frequency axis. The default is `"auto"` which selects `"one-sided"`
            for real-valued filters and `"two-sided"` for complex-valued filters.
        decades: The number of decades to plot when `x_axis="log"`.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

    Examples:
        See the :ref:`fir-filters` example.

        .. ipython:: python

            h_srrc = sdr.root_raised_cosine(0.5, 10, 10)

            @savefig sdr_plot_phase_response_1.png
            plt.figure(); \
            sdr.plot.phase_response(h_srrc)

        See the :ref:`iir-filters` example.

        .. ipython:: python

            zero = 0.6; \
            pole = 0.8 * np.exp(1j * np.pi / 8); \
            iir = sdr.IIR.ZerosPoles([zero], [pole, pole.conj()])

            @savefig sdr_plot_phase_response_2.png
            plt.figure(); \
            sdr.plot.phase_response(iir)

        .. ipython:: python

            @savefig sdr_plot_phase_response_3.png
            plt.figure(); \
            sdr.plot.phase_response(h_srrc, x_axis="two-sided")

        .. ipython:: python

            @savefig sdr_plot_phase_response_4.png
            plt.figure(); \
            sdr.plot.phase_response(iir, x_axis="log", decades=3)

    Group:
        plot-filter
    """
    b, a = _convert_to_taps(filter)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate)
    verify_scalar(N, optional=True, int=True, positive=True)
    verify_bool(unwrap)
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(x_axis, ["auto", "one-sided", "two-sided", "log"])
    verify_scalar(decades, int=True, positive=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if x_axis == "auto":
            x_axis = "one-sided" if np.isrealobj(b) and np.isrealobj(a) else "two-sided"

        if x_axis == "log":
            f = np.logspace(np.log10(sample_rate / 2 / 10**decades), np.log10(sample_rate / 2), N)
            f, H = scipy.signal.freqz(b, a, worN=f, whole=False, fs=sample_rate)
        else:
            f, H = scipy.signal.freqz(b, a, worN=N, whole=x_axis == "two-sided", fs=sample_rate)

        if x_axis == "two-sided":
            f -= sample_rate / 2
            H = np.fft.fftshift(H)

        if sample_rate_provided:
            units, scalar = freq_units(f)
            f *= scalar

        if unwrap:
            theta = np.rad2deg(np.unwrap(np.angle(H)))
        else:
            theta = np.rad2deg(np.angle(H))

        if x_axis == "two-sided":
            # Set omega=0 to have phase of 0
            theta -= theta[np.argmin(np.abs(f))]

        if x_axis == "log":
            ax.semilogx(f, theta, **kwargs)
        else:
            ax.plot(f, theta, **kwargs)

        ax.grid(True, which="both")
        if "label" in kwargs:
            ax.legend()
        if sample_rate_provided:
            ax.set_xlabel(f"Frequency ({units}), $f$")
        else:
            ax.set_xlabel("Normalized frequency, $f / f_s$")
        ax.set_ylabel(r"Phase (deg), $\angle H(\omega)$")
        ax.set_title(r"Phase response, $\angle H(\omega)$")


@export
def phase_delay(
    filter: FIR | IIR | npt.ArrayLike | tuple[npt.ArrayLike, npt.ArrayLike],
    sample_rate: float | None = None,
    N: int = 1024,
    ax: plt.Axes | None = None,
    x_axis: Literal["auto", "one-sided", "two-sided", "log"] = "auto",
    decades: int = 4,
    **kwargs,
):
    r"""
    Plots the phase delay $\tau_{\phi}(\omega)$ of the filter.

    Arguments:
        filter: The filter definition.

            - :class:`sdr.FIR`, :class:`sdr.IIR`: The filter object.
            - `npt.ArrayLike`: The feedforward coefficients $b_i$.
            - `tuple[npt.ArrayLike, npt.ArrayLike]`: The feedforward coefficients $b_i$ and
              feedback coefficients $a_j$.

        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Normalized frequency".
        N: The number of samples $N$ in the phase delay.
        ax: The axis to plot on. If `None`, the current axis is used.
        x_axis: The x-axis scaling. Options are to display a one-sided spectrum, a two-sided spectrum, or
            one-sided spectrum with a logarithmic frequency axis. The default is `"auto"` which selects `"one-sided"`
            for real-valued filters and `"two-sided"` for complex-valued filters.
        decades: The number of decades to plot when `x_axis="log"`.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

    Examples:
        See the :ref:`fir-filters` example.

        .. ipython:: python

            h_srrc = sdr.root_raised_cosine(0.5, 10, 10)

            @savefig sdr_plot_phase_delay_1.png
            plt.figure(); \
            sdr.plot.phase_delay(h_srrc)

        See the :ref:`iir-filters` example.

        .. ipython:: python

            zero = 0.6; \
            pole = 0.8 * np.exp(1j * np.pi / 8); \
            iir = sdr.IIR.ZerosPoles([zero], [pole, pole.conj()])

            @savefig sdr_plot_phase_delay_2.png
            plt.figure(); \
            sdr.plot.phase_delay(iir)

        .. ipython:: python

            @savefig sdr_plot_phase_delay_3.png
            plt.figure(); \
            sdr.plot.phase_delay(h_srrc, x_axis="two-sided")

        .. ipython:: python

            @savefig sdr_plot_phase_delay_4.png
            plt.figure(); \
            sdr.plot.phase_delay(iir, x_axis="log", decades=3)

    Group:
        plot-filter
    """
    b, a = _convert_to_taps(filter)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate)
    verify_scalar(N, optional=True, int=True, positive=True)
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(x_axis, ["auto", "one-sided", "two-sided", "log"])
    verify_scalar(decades, int=True, positive=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if x_axis == "auto":
            x_axis = "one-sided" if np.isrealobj(b) and np.isrealobj(a) else "two-sided"

        if x_axis == "log":
            f = np.logspace(np.log10(sample_rate / 2 / 10**decades), np.log10(sample_rate / 2), N)
            f, H = scipy.signal.freqz(b, a, worN=f, whole=False, fs=sample_rate)
        else:
            f, H = scipy.signal.freqz(b, a, worN=N, whole=x_axis == "two-sided", fs=sample_rate)

        if x_axis == "two-sided":
            f -= sample_rate / 2
            H = np.fft.fftshift(H)

        theta = np.unwrap(np.angle(H))
        theta -= theta[np.argmin(np.abs(f))]  # Set omega=0 to have phase of 0
        with np.errstate(divide="ignore", invalid="ignore"):
            tau_phi = -theta / (2 * np.pi * f)
        tau_phi[np.argmin(np.abs(f))] = np.nan  # Avoid crazy result when dividing by near zero

        if sample_rate_provided:
            f_units, scalar = freq_units(f)
            f *= scalar
            t_units, scalar = time_units(tau_phi)
            tau_phi *= scalar

        if x_axis == "log":
            ax.semilogx(f, tau_phi, **kwargs)
        else:
            ax.plot(f, tau_phi, **kwargs)

        min_ylim(tau_phi, 2 / sample_rate, sample_rate)

        ax.grid(True, which="both")
        if "label" in kwargs:
            ax.legend()
        if sample_rate_provided:
            ax.set_xlabel(f"Frequency ({f_units}), $f$")
            ax.set_ylabel(rf"Phase delay ({t_units}), $\tau_{{\phi}}{{\omega}}$")
        else:
            ax.set_xlabel("Normalized frequency, $f / f_s$")
            ax.set_ylabel(r"Phase delay (samples), $\tau_{\phi}(\omega)$")
        ax.set_title(r"Phase delay, $\tau_{\phi}(\omega)$")


@export
def group_delay(
    filter: FIR | IIR | npt.ArrayLike | tuple[npt.ArrayLike, npt.ArrayLike],
    sample_rate: float | None = None,
    N: int = 1024,
    ax: plt.Axes | None = None,
    x_axis: Literal["auto", "one-sided", "two-sided", "log"] = "auto",
    decades: int = 4,
    **kwargs,
):
    r"""
    Plots the group delay $\tau_g(\omega)$ of the IIR filter.

    Arguments:
        filter: The filter definition.

            - :class:`sdr.FIR`, :class:`sdr.IIR`: The filter object.
            - `npt.ArrayLike`: The feedforward coefficients $b_i$.
            - `tuple[npt.ArrayLike, npt.ArrayLike]`: The feedforward coefficients $b_i$ and
              feedback coefficients $a_j$.

        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Normalized frequency".
        N: The number of samples $N$ in the frequency response.
        ax: The axis to plot on. If `None`, the current axis is used.
        x_axis: The x-axis scaling. Options are to display a one-sided spectrum, a two-sided spectrum, or
            one-sided spectrum with a logarithmic frequency axis. The default is `"auto"` which selects `"one-sided"`
            for real-valued filters and `"two-sided"` for complex-valued filters.
        decades: The number of decades to plot when `x_axis="log"`.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

    Examples:
        See the :ref:`fir-filters` example.

        .. ipython:: python

            h_srrc = sdr.root_raised_cosine(0.5, 10, 10)

            @savefig sdr_plot_group_delay_1.png
            plt.figure(); \
            sdr.plot.group_delay(h_srrc);

        See the :ref:`iir-filters` example.

        .. ipython:: python

            zero = 0.6; \
            pole = 0.8 * np.exp(1j * np.pi / 8); \
            iir = sdr.IIR.ZerosPoles([zero], [pole, pole.conj()])

            @savefig sdr_plot_group_delay_2.png
            plt.figure(); \
            sdr.plot.group_delay(iir)

        .. ipython:: python

            @savefig sdr_plot_group_delay_3.png
            plt.figure(); \
            sdr.plot.group_delay(h_srrc, x_axis="two-sided");

        .. ipython:: python

            @savefig sdr_plot_group_delay_4.png
            plt.figure(); \
            sdr.plot.group_delay(iir, x_axis="log", decades=3)

    Group:
        plot-filter
    """
    b, a = _convert_to_taps(filter)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate)
    verify_scalar(N, optional=True, int=True, positive=True)
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(x_axis, ["auto", "one-sided", "two-sided", "log"])
    verify_scalar(decades, int=True, positive=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if x_axis == "auto":
            x_axis = "one-sided" if np.isrealobj(b) and np.isrealobj(a) else "two-sided"

        if x_axis == "log":
            f = np.logspace(np.log10(sample_rate / 2 / 10**decades), np.log10(sample_rate / 2), N)
            f, tau_g = scipy.signal.group_delay((b, a), w=f, whole=False, fs=sample_rate)
        else:
            f, tau_g = scipy.signal.group_delay((b, a), w=N, whole=x_axis == "two-sided", fs=sample_rate)

        if x_axis == "two-sided":
            f -= sample_rate / 2
            tau_g = np.fft.fftshift(tau_g)

        if sample_rate_provided:
            f_units, scalar = freq_units(f)
            f *= scalar

            tau_g /= sample_rate
            t_units, scalar = time_units(tau_g)
            tau_g *= scalar

        if x_axis == "log":
            ax.semilogx(f, tau_g, **kwargs)
        else:
            ax.plot(f, tau_g, **kwargs)

        min_ylim(tau_g, 2 / sample_rate, sample_rate)

        ax.grid(True, which="both")
        if "label" in kwargs:
            ax.legend()
        if sample_rate_provided:
            ax.set_xlabel(f"Frequency ({f_units}), $f$")
            ax.set_ylabel(rf"Group delay ({t_units}), $\tau_g(\omega)$")
        else:
            ax.set_xlabel("Normalized frequency, $f / f_s$")
            ax.set_ylabel(r"Group delay (samples), $\tau_g(\omega)$")
        ax.set_title(r"Group delay, $\tau_g(\omega)$")


@export
def filter(
    filter: FIR | IIR | npt.ArrayLike | tuple[npt.ArrayLike, npt.ArrayLike],
    sample_rate: float | None = None,
    N_time: int | None = None,
    N_freq: int = 1024,
    x_axis: Literal["one-sided", "two-sided", "log"] = "two-sided",
    decades: int = 4,
):
    r"""
    Plots the magnitude response $|H(\omega)|^2$, impulse response $h[n]$, and zeros and poles of the filter.

    Arguments:
        filter: The filter definition.

            - :class:`sdr.FIR`, :class:`sdr.IIR`: The filter object.
            - `npt.ArrayLike`: The feedforward coefficients $b_i$.
            - `tuple[npt.ArrayLike, npt.ArrayLike]`: The feedforward coefficients $b_i$ and
              feedback coefficients $a_j$.

        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Normalized frequency".
        N_time: The number of samples $N_t$ in the time domain. If `None`, the length of `b` is used
            for FIR filters and 100 for IIR filters.
        N_freq: The number of samples $N_f$ in the frequency response.
        x_axis: The x-axis scaling. Options are to display a one-sided spectrum, a two-sided spectrum, or
            one-sided spectrum with a logarithmic frequency axis.
        decades: The number of decades to plot when `x_axis="log"`.

    Examples:
        See the :ref:`fir-filters` example.

        .. ipython:: python

            h_srrc = sdr.root_raised_cosine(0.5, 10, 10)

            @savefig sdr_plot_filter_1.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.filter(h_srrc)

        See the :ref:`iir-filters` example.

        .. ipython:: python

            zero = 0.6; \
            pole = 0.8 * np.exp(1j * np.pi / 8); \
            iir = sdr.IIR.ZerosPoles([zero], [pole, pole.conj()])

            @savefig sdr_plot_filter_2.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.filter(iir, N_time=30)

    Group:
        plot-filter
    """
    with plt.rc_context(RC_PARAMS):
        b, a = _convert_to_taps(filter)

        plt.subplot2grid((2, 3), (0, 0), 1, 3)
        magnitude_response((b, a), sample_rate=sample_rate, N=N_freq, x_axis=x_axis, decades=decades)

        plt.subplot2grid((2, 3), (1, 0), 1, 1)
        zeros_poles((b, a))

        plt.subplot2grid((2, 3), (1, 1), 1, 2)
        impulse_response((b, a), N=N_time)



"""
A module containing functions for calculating path losses.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.constants

from .._conversion import db
from .._helper import convert_output, export, verify_arraylike


@export
def free_space_path_loss(
    distance: npt.ArrayLike,
    freq: npt.ArrayLike,
) -> npt.NDArray[np.float64]:
    r"""
    Calculates the free-space path loss (FSPL) in dB.

    $$\text{FSPL} = 10 \log_{10} \left( \frac{4 \pi d f}{c} \right)^2$$

    Arguments:
        distance: The distance $d$ in meters between the transmitter and receiver.
        freq: The frequency $f$ in Hz of the signal.

    Returns:
        The free-space path loss (FSPL) in dB.

    Note:
        The free-space path loss equation is only valid in the far field. For $d < \lambda / 4 \pi$, the FSPL
        equation has a negative result, implying a path gain. This is not possible. So these path losses
        are set to 0 dB.

    Examples:
        Compute the free-space path loss for a 1 km link at 1 GHz.

        .. ipython:: python

            sdr.free_space_path_loss(1e3, 1e9)

        The free-space path loss is proportional to the square of the distance. So, doubling the distance
        results in a 6 dB increase in the free-space path loss.

        .. ipython:: python

            sdr.free_space_path_loss(2e3, 1e9)

        The free-space path loss is also proportional to the square of the frequency. So, doubling the frequency
        results in a 6 dB increase in the free-space path loss.

        .. ipython:: python

            sdr.free_space_path_loss(1e3, 2e9)

        Plot the free-space path loss at 1 GHz for distances up to 1 km.

        .. ipython:: python

            d = np.linspace(0, 1_000, 1000)  # m
            fspl = sdr.free_space_path_loss(d, 1e9)  # dB

            @savefig sdr_fspl_1.png
            plt.figure(); \
            plt.plot(d, fspl); \
            plt.xlabel('Distance (m)'); \
            plt.ylabel('Free-space path loss (dB)');

        It is confusing why free-space path loss is proportional to the square of frequency. Is RF energy attenuated
        more at higher frequencies? The answer is no. The reason the FSPL equation has a frequency
        dependence is that it assumes omnidirectional, isotropic antennas are used at both the transmitter and
        receiver. The isotropic antenna has a gain of 0 dBi. The physical size of a roughly isotropic antenna
        (think dipole) is a function of frequency as well. So, as the frequency increases, the physical size of the
        isotropic antenna decreases. But what if the size of the antenna was fixed across frequency, as is the case
        with a parabolic dish antenna? You'll note that the gain of a parabolic dish antenna is also proportional to
        the square of the frequency, see :func:`sdr.parabolic_antenna`.

        It turns out that if omnidirectional antennas are used at both the transmitter and receiver, the total path
        loss increases with frequency. But if a parabolic reflector is used at one end, the total path loss is
        constant across frequency. Furthermore, if a parabolic reflector is used at both ends, as is the case in
        VSAT systems, the total path loss decreases with frequency.

        .. ipython:: python

            freq = np.linspace(1e6, 40e9, 1_001)

            # Free-space path loss at 1 km
            fspl = sdr.free_space_path_loss(1e3, freq)

            # Isotropic antenna gain in dBi
            iso = 0

            # 1-meter diameter parabolic dish antenna gain in dBi
            par = sdr.parabolic_antenna(freq, 1)[0]

            @savefig sdr_fspl_2.png
            plt.figure(); \
            plt.plot(freq / 1e9, fspl - iso - iso, label="Isotropic -> Isotropic"); \
            plt.plot(freq / 1e9, fspl - iso - par, label="Isotropic -> Parabolic"); \
            plt.plot(freq / 1e9, fspl - par - par, label="Parabolic -> Parabolic"); \
            plt.legend(title="Antennas", loc="center right"); \
            plt.xlabel("Frequency (GHz), $f$"); \
            plt.ylabel("Path loss (dB)"); \
            plt.title("Path loss across center frequency");

    Group:
        link-budget-path-losses
    """
    distance = verify_arraylike(distance, float=True, non_negative=True)
    freq = verify_arraylike(freq, float=True, positive=True)

    # The free-space path loss equation is only valid in the far field. For very small distances, the FSPL
    # equation could have a negative result, implying a path gain. This is not possible. So for distances less
    # than lambda / (4 * pi), the path loss is set to 0 dB.
    lambda_ = scipy.constants.speed_of_light / freq
    distance = np.maximum(distance, lambda_ / (4 * np.pi))

    # The free-space path loss equation
    loss = db((4 * np.pi * distance * freq / scipy.constants.speed_of_light) ** 2)

    return convert_output(loss)



import pytest

import sdr


def test_stein_1():
    snr1 = -10  # dB
    snr2 = -10  # dB
    time = 1e-3  # seconds
    bandwidth = 1e6  # Hz
    sigma_fdoa = sdr.fdoa_crlb(snr1, snr2, time, bandwidth)
    assert sigma_fdoa == pytest.approx(135.04744742356585)  # Stein says ~140 Hz


def test_stein_2():
    snr1 = -10  # dB
    snr2 = -10  # dB
    time = 6e-3  # seconds
    bandwidth = 1e6  # Hz
    sigma_fdoa = sdr.fdoa_crlb(snr1, snr2, time, bandwidth)
    assert sigma_fdoa == pytest.approx(9.188814923696532)  # Stein says ~9 ns



"""
A module containing time-domain plotting functions.
"""

from __future__ import annotations

from typing import overload

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib.collections import LineCollection
from typing_extensions import Literal

from .._conversion import db
from .._helper import (
    export,
    verify_arraylike,
    verify_bool,
    verify_isinstance,
    verify_literal,
    verify_not_specified,
    verify_positional_args,
    verify_scalar,
    verify_specified,
)
from ._helper import integer_x_axis, real_or_complex_plot, verify_sample_rate
from ._rc_params import RC_PARAMS
from ._units import time_units


@overload
def time_domain(
    x: npt.NDArray,  # TODO: Change to npt.ArrayLike once Sphinx has better overload support
    *,
    sample_rate: float | None = None,
    centered: bool = False,
    offset: float = 0.0,
    ax: plt.Axes | None = None,
    diff: Literal["color", "line"] = "color",
    **kwargs,
): ...


@overload
def time_domain(
    t: npt.NDArray,  # TODO: Change to npt.ArrayLike once Sphinx has better overload support
    x: npt.NDArray,  # TODO: Change to npt.ArrayLike once Sphinx has better overload support
    *,
    sample_rate: float | None = None,
    centered: bool = False,
    offset: float = 0.0,
    ax: plt.Axes | None = None,
    diff: Literal["color", "line"] = "color",
    **kwargs,
): ...


@export
def time_domain(  # noqa: D417
    *args,
    sample_rate: float | None = None,
    centered: bool = False,
    offset: float = 0.0,
    ax: plt.Axes | None = None,
    diff: Literal["color", "line"] = "color",
    **kwargs,
):
    r"""
    Plots a time-domain signal $x[n]$.

    Arguments:
        t: The time signal $t[n]$. The units are assumed to be $1/f_s$.
        x: The time-domain signal $x[n]$.
        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Samples".
        centered: Indicates whether to center the x-axis about 0. This argument is mutually exclusive with
            `offset`.
        offset: The x-axis offset to apply to the first sample. The units of the offset are $1/f_s$.
            This argument is mutually exclusive with `centered`.
        ax: The axis to plot on. If `None`, the current axis is used.
        diff: Indicates how to differentiate the real and imaginary parts of a complex signal. If `"color"`, the
            real and imaginary parts will have different colors based on the current Matplotlib color cycle.
            If `"line"`, the real part will have a solid line and the imaginary part will have a dashed line,
            and both lines will share the same color.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

    Examples:
        Plot a square-root raised cosine (SRRC) pulse shape centered about 0.

        .. ipython:: python

            qpsk = sdr.PSK(4, phase_offset=45, sps=10, pulse_shape="srrc"); \
            pulse_shape = qpsk.pulse_shape

            @savefig sdr_plot_time_domain_1.png
            plt.figure(); \
            sdr.plot.time_domain(pulse_shape, centered=True); \
            plt.title("SRRC pulse shape");

        Plot an imaginary QPSK signal at 10 kS/s.

        .. ipython:: python

            symbols = np.random.randint(0, 4, 50); \
            x = qpsk.modulate(symbols)

            @savefig sdr_plot_time_domain_2.png
            plt.figure(); \
            sdr.plot.time_domain(x, sample_rate=10e3); \
            plt.title("SRRC pulse-shaped QPSK");

        Plot non-uniformly sampled data.

        .. ipython:: python

            t = np.array([0, 1, 2, 3, 5, 8, 13, 21, 34, 55]); \
            x = np.random.randn(t.size)

            @savefig sdr_plot_time_domain_3.png
            plt.figure(); \
            sdr.plot.time_domain(t, x, marker="."); \
            plt.title("Non-uniformly sampled data");

    Group:
        plot-time-domain
    """
    verify_positional_args(args, 2)
    if len(args) == 1:
        x = verify_arraylike(args[0], complex=True, ndim=1)
        t = None
    elif len(args) == 2:
        t = verify_arraylike(args[0], float=True, ndim=1)
        x = verify_arraylike(args[1], complex=True, ndim=1)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate)
    verify_bool(centered)
    verify_scalar(offset, float=True)
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(diff, ["color", "line"])

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if t is None:
            if centered:
                if x.size % 2 == 0:
                    t = np.arange(-x.size // 2, x.size // 2) / sample_rate
                else:
                    t = np.arange(-(x.size - 1) // 2, (x.size + 1) // 2) / sample_rate
            else:
                t = np.arange(x.size) / sample_rate + offset

        if sample_rate_provided:
            units, scalar = time_units(t)
            t *= scalar

        real_or_complex_plot(t, x, ax=ax, diff=diff, **kwargs)
        if sample_rate_provided:
            ax.set_xlabel(f"Time ({units})")
        else:
            integer_x_axis(ax)
            ax.set_xlabel("Sample, $n$")
        ax.set_ylabel("Amplitude")


@export
def raster(
    x: npt.ArrayLike,
    length: int | None = None,
    stride: int | None = None,
    sample_rate: float | None = None,
    color: Literal["index"] | str = "index",
    persistence: bool = False,
    colorbar: bool = True,
    ax: plt.Axes | None = None,
    **kwargs,
):
    r"""
    Plots a raster of the time-domain signal $x[n]$.

    Arguments:
        x: The real time-domain signal $x[n]$. If `x` is 1D, the rastering is determined by `length` and `stride`.
            If `x` is 2D, the rows correspond to each raster.
        length: The length of each raster in samples. This must be provided if `x` is 1D.
        stride: The stride between each raster in samples. If `None`, the stride is set to `length`.
        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Samples".
        color: Indicates how to color the rasters. If `"index"`, the rasters are colored based on their index.
            If a valid Matplotlib color, the rasters are all colored with that color.
        persistence: Indicates whether to plot the raster as a persistence plot. A persistence plot is a
            2D histogram of the rasters.
        colorbar: Indicates whether to add a colorbar to the plot. This is only added if `color="index"` or
            `persistence=True`.
        ax: The axis to plot on. If `None`, the current axis is used.
        kwargs: Additional keyword arguments to pass to Matplotlib functions.

            If `persistence=False`, the following keyword arguments are passed to
            :obj:`matplotlib.collections.LineCollection`. The defaults may be overwritten.

            - `"linewidths"`: 1
            - `"linestyles"`: `"solid"`
            - `"cmap"`: `"rainbow"`

            If `persistence=True`, the following keyword arguments are passed to :func:`matplotlib.pyplot.pcolormesh`.
            The defaults may be overwritten.

            - `"bins"`: `(800, 200)  # Passed to np.histogram2d()`
            - `"cmap"`: `"rainbow"`
            - `"norm"`: `"log"`
            - `"rasterized"`: `True`
            - `"show_zero"`: `False`

    Group:
        plot-time-domain
    """
    x = verify_arraylike(x, float=True)  # TODO: Check ndim in [1, 2]
    if not x.ndim in [1, 2]:
        raise ValueError(f"Argument 'x' must be 1-D or 2-D, not {x.ndim}-D.")
    verify_scalar(length, optional=True, int=True, inclusive_min=1, inclusive_max=x.size)
    verify_scalar(stride, optional=True, int=True, inclusive_min=1, inclusive_max=x.size)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate)
    verify_isinstance(color, str)
    verify_bool(persistence)
    verify_bool(colorbar)
    verify_isinstance(ax, plt.Axes, optional=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if x.ndim == 1:
            verify_specified(length)
            if stride is None:
                stride = length

            # Compute the strided data and format into segments for LineCollection
            N_rasters = (x.size - length) // stride + 1
            x_strided = np.lib.stride_tricks.as_strided(
                x, shape=(N_rasters, length), strides=(x.strides[0] * stride, x.strides[0]), writeable=False
            )
        else:
            verify_not_specified(length)
            verify_not_specified(stride)

            N_rasters, length = x.shape
            x_strided = x

        t = np.arange(length) / sample_rate
        if sample_rate_provided:
            units, scalar = time_units(t)
            t *= scalar

        if persistence:
            default_kwargs = {
                "bins": (800, 200),
                "cmap": "rainbow",
                "norm": "log",
                "rasterized": True,
                "show_zero": False,
            }
            kwargs = {**default_kwargs, **kwargs}

            bins = kwargs.pop("bins")
            t_fine = np.linspace(t.min(), t.max(), bins[0])
            x_fine = np.concatenate([np.interp(t_fine, t, x_row) for x_row in x_strided])
            t_fine = np.broadcast_to(t_fine, (x_strided.shape[0], bins[0])).ravel()
            h, t_edges, x_edges = np.histogram2d(t_fine, x_fine, bins=bins)

            cmap = kwargs.pop("cmap")  # Need to pop cmap to avoid passing it twice to pcolormesh
            cmap = plt.colormaps[cmap]
            show_zero = kwargs.pop("show_zero")
            if show_zero:
                cmap = cmap.with_extremes(bad=cmap(0))
            else:
                h[h == 0] = np.nan  # Set 0s to NaNs so they don't show up in the plot

            pcm = ax.pcolormesh(t_edges, x_edges, h.T, cmap=cmap, **kwargs)
            if colorbar:
                plt.colorbar(pcm, label="Points", pad=0.05)
        else:
            segments = [np.column_stack([t, x_raster]) for x_raster in x_strided]

            # Set the default keyword arguments and override with user-specified keyword arguments
            default_kwargs = {
                "linewidths": 1,
                "linestyles": "solid",
                "cmap": "rainbow",
            }
            if color == "index":
                default_kwargs["array"] = np.arange(N_rasters)
            else:
                default_kwargs["colors"] = color
            kwargs = {**default_kwargs, **kwargs}

            line_collection = LineCollection(
                segments,
                **kwargs,
            )

            ax.add_collection(line_collection)
            ax.set_xlim(t.min(), t.max())
            ax.set_ylim(x.min(), x.max())

            if colorbar and color == "index":
                plt.colorbar(line_collection, label="Raster Index", pad=0.05)

        if sample_rate_provided:
            ax.set_xlabel(f"Time ({units})")
        else:
            integer_x_axis(ax)
            ax.set_xlabel("Sample, $n$")
        ax.set_ylabel("Amplitude")


@export
def correlation(
    x: npt.ArrayLike,
    y: npt.ArrayLike,
    sample_rate: float | None = None,
    mode: Literal["full", "valid", "same", "circular"] = "full",
    ax: plt.Axes | None = None,
    y_axis: Literal["complex", "mag", "mag^2", "db"] = "mag",
    diff: Literal["color", "line"] = "color",
    **kwargs,
):
    r"""
    Plots the correlation between two time-domain signals $x[n]$ and $y[n]$.

    Arguments:
        x: The first time-domain signal $x[n]$.
        y: The second time-domain signal $y[n]$.
        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Lag (samples)".
        mode: The :func:`numpy.correlate` correlation mode. If `"circular"`, a circular correlation is computed
            using FFTs.
        ax: The axis to plot on. If `None`, the current axis is used.
        y_axis: Indicates how to plot the y-axis. If `"complex"`, the real and imaginary parts are plotted separately.
        diff: Indicates how to differentiate the real and imaginary parts of a complex signal. If `"color"`, the
            real and imaginary parts will have different colors based on the current Matplotlib color cycle.
            If `"line"`, the real part will have a solid line and the imaginary part will have a dashed line,
            and both lines will share the same color.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

    Examples:
        Plot the auto-correlation of a length-63 $m$-sequence. Notice that the linear correlation produces sidelobes
        for non-zero lag.

        .. ipython:: python

            x = sdr.m_sequence(6, output="bipolar")

            @savefig sdr_plot_correlation_1.png
            plt.figure(); \
            sdr.plot.correlation(x, x, mode="full");

        However, the circular correlation only produces magnitudes of 1 for non-zero lag.

        .. ipython:: python

            @savefig sdr_plot_correlation_2.png
            plt.figure(); \
            sdr.plot.correlation(x, x, mode="circular");

    Group:
        plot-time-domain
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    y = verify_arraylike(y, complex=True, ndim=1)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate)
    verify_literal(mode, ["full", "valid", "same", "circular"])
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(y_axis, ["complex", "mag", "mag^2", "db"])
    verify_literal(diff, ["color", "line"])

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if mode == "circular":
            n_fft = max(x.size, y.size)
            X = np.fft.fft(x, n_fft)
            Y = np.fft.fft(y, n_fft)
            corr = np.fft.ifft(X * Y.conj(), n_fft)
            corr = np.fft.fftshift(corr)
            if np.isrealobj(x) and np.isrealobj(y):
                # If both signals are real, the correlation is real
                corr = corr.real
        else:
            corr = np.correlate(x, y, mode=mode)

        if corr.size % 2 == 0:
            t = np.arange(-corr.size // 2, corr.size // 2)  # Lags
        else:
            t = np.arange(-corr.size // 2 + 1, corr.size // 2 + 1)  # Lags

        if sample_rate_provided:
            units, scalar = time_units(t)
            t *= scalar

        if x is y:
            equation = r"r_{xx}[n]"
            if mode == "circular":
                ax.set_title("Periodic auto-correlation function (PACF)")
            else:
                ax.set_title("Auto-correlation function (ACF)")
        else:
            equation = r"r_{xy}[n]"
            if mode == "circular":
                ax.set_title("Periodic cross-correlation function (PCCF)")
            else:
                ax.set_title("Cross-correlation function (CCF)")

        if y_axis == "complex":
            ax.set_ylabel(rf"Correlation, ${equation}$")
        elif y_axis == "mag":
            corr = np.abs(corr)
            ax.set_ylabel(rf"Correlation, $\left| {equation} \right|$")
        elif y_axis == "mag^2":
            corr = np.abs(corr) ** 2
            ax.set_ylabel(rf"Correlation, $\left| {equation} \right|^2$")
        elif y_axis == "db":
            corr = db(np.abs(corr) ** 2)
            ax.set_ylabel(rf"Correlation (dB), $\left| {equation} \right|^2$")

        real_or_complex_plot(t, corr, ax=ax, diff=diff, **kwargs)
        if sample_rate_provided:
            ax.set_xlabel(rf"Lag ({units}), $\Delta t$")
        else:
            integer_x_axis(ax)
            ax.set_xlabel(r"Lag (samples), $\Delta n$")



import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

import sdr


def debug_plot(h: npt.NDArray, h_truth: npt.NDArray):
    plt.figure()
    sdr.plot.time_domain(h_truth, marker=".", label="h_truth")
    sdr.plot.time_domain(h, marker="x", label="h")
    plt.legend()

    plt.figure()
    sdr.plot.magnitude_response(h_truth, label="h_truth")
    sdr.plot.magnitude_response(h, label="h")
    plt.legend()

    plt.show()


def verify_impulse_response(h: npt.NDArray, h_truth: npt.NDArray, atol: float = 1e-6):
    # debug_plot(h, h_truth)

    # MATLAB sets the center of the passband to 0 dB gain. We set the average passband gain to 0 dB.
    # Normalize the impulse response to have the same RMS voltage as MATLAB for comparison.
    h *= sdr.rms_voltage(h_truth) / sdr.rms_voltage(h)

    assert h.shape == h_truth.shape
    assert np.allclose(h, h_truth, atol=atol)



import numpy as np

import sdr


def test_freq():
    sample_rate = 1e3  # samples/s
    freq = 10  # Hz
    N = 100  # samples
    x = np.exp(1j * (2 * np.pi * freq / sample_rate * np.arange(N)))
    y = sdr.mix(x, freq=-freq, sample_rate=sample_rate)
    assert np.allclose(y, 1)


def test_phase():
    sample_rate = 1e3  # samples/s
    phase = 45  # degrees
    N = 100  # samples
    x = np.exp(1j * np.ones(N) * np.deg2rad(phase))
    y = sdr.mix(x, phase=-phase, sample_rate=sample_rate)
    assert np.allclose(y, 1)


def test_freq_phase():
    sample_rate = 1e3  # samples/s
    freq = 10  # Hz
    phase = 45  # degrees
    N = 100  # samples
    x = np.exp(1j * (2 * np.pi * freq / sample_rate * np.arange(N) + np.deg2rad(phase)))
    y = sdr.mix(x, freq=-freq, phase=-phase, sample_rate=sample_rate)
    assert np.allclose(y, 1)



import numpy as np
import pytest

import sdr


@pytest.mark.parametrize(
    "pole", [0.8, 0.8 * np.exp(1j * np.pi / 4), 1.2, 1.2 * np.exp(1j * np.pi / 4), 1.2, 1.2 * np.exp(1j * np.pi / 4)]
)
def test_single_pole_impulse_response(pole):
    """
    Reference:
        - R. Lyons, Understanding Digital Signal Processing 3rd Edition, Section 6.3.1.
    """
    N = 100
    x = np.zeros(N, dtype=float)
    x[0] = 1

    b = np.array([1], dtype=complex)
    a = np.array([1, -pole], dtype=complex)
    iir = sdr.IIR(b, a)
    h = iir(x)
    h_truth = pole ** np.arange(N)

    assert h.shape == h_truth.shape
    assert np.allclose(h, h_truth)


def test_streaming_match_full():
    N = 50
    rng = np.random.default_rng()
    x = rng.standard_normal(N) + 1j * rng.standard_normal(N)  # Input signal

    # Define random zeros and poles
    Nz = rng.integers(1, 5)
    zeros = rng.uniform(0.5, 1, size=Nz) * np.exp(1j * rng.uniform(-np.pi, np.pi, size=Nz))
    Np = rng.integers(1, 5)
    poles = rng.uniform(0.2, 0.8, size=Np) * np.exp(1j * rng.uniform(-np.pi, np.pi, size=Np))

    iir1 = sdr.IIR.ZerosPoles(zeros, poles)
    y_full = iir1(x)

    iir2 = sdr.IIR.ZerosPoles(zeros, poles, streaming=True)
    d = 10  # Stride
    y_stream = np.zeros_like(y_full)
    for i in range(0, N, d):
        y_stream[i : i + d] = iir2(x[i : i + d])

    np.testing.assert_array_almost_equal(y_full, y_stream)



"""
References:
    - https://www.gaussianwaves.com/2020/09/equivalent-noise-bandwidth-enbw-of-window-functions/
"""

import pytest
import scipy.signal

import sdr


def test_boxcar():
    n = 128
    h = scipy.signal.windows.get_window("boxcar", n)
    fir = sdr.FIR(h)
    B_n = fir.noise_bandwidth(n)
    assert B_n == pytest.approx(1.000, rel=1e-3)


def test_barthann():
    n = 128
    h = scipy.signal.windows.get_window("barthann", n)
    fir = sdr.FIR(h)
    B_n = fir.noise_bandwidth(n)
    assert B_n == pytest.approx(1.456, rel=1e-3)


def test_bartlett():
    n = 128
    h = scipy.signal.windows.get_window("bartlett", n)
    fir = sdr.FIR(h)
    B_n = fir.noise_bandwidth(n)
    assert B_n == pytest.approx(1.333, rel=1e-3)


def test_blackman():
    n = 128
    h = scipy.signal.windows.get_window("blackman", n)
    fir = sdr.FIR(h)
    B_n = fir.noise_bandwidth(n)
    assert B_n == pytest.approx(1.727, rel=1e-3)


def test_blackmanharris():
    n = 128
    h = scipy.signal.windows.get_window("blackmanharris", n)
    fir = sdr.FIR(h)
    B_n = fir.noise_bandwidth(n)
    assert B_n == pytest.approx(2.004, rel=1e-3)


def test_bohman():
    n = 128
    h = scipy.signal.windows.get_window("bohman", n)
    fir = sdr.FIR(h)
    B_n = fir.noise_bandwidth(n)
    assert B_n == pytest.approx(1.786, rel=1e-3)


def test_flattop():
    n = 128
    h = scipy.signal.windows.get_window("flattop", n)
    fir = sdr.FIR(h)
    B_n = fir.noise_bandwidth(n)
    assert B_n == pytest.approx(3.770, rel=1e-3)


def test_hamming():
    n = 128
    h = scipy.signal.windows.get_window("hamming", n)
    fir = sdr.FIR(h)
    B_n = fir.noise_bandwidth(n)
    assert B_n == pytest.approx(1.363, rel=1e-3)


def test_hann():
    n = 128
    h = scipy.signal.windows.get_window("hann", n)
    fir = sdr.FIR(h)
    B_n = fir.noise_bandwidth(n)
    assert B_n == pytest.approx(1.500, rel=1e-3)


def test_nuttall():
    n = 128
    h = scipy.signal.windows.get_window("nuttall", n)
    fir = sdr.FIR(h)
    B_n = fir.noise_bandwidth(n)
    assert B_n == pytest.approx(1.976, rel=1e-3)



"""
References:
    - https://en.wikipedia.org/wiki/Crest_factor
"""

import numpy as np
import pytest

import sdr


def test_sine():
    N = 50
    x = np.sin(np.linspace(0, 2 * np.pi, N, endpoint=False))
    assert sdr.crest_factor(x) == pytest.approx(np.sqrt(2), rel=1e-1)


def test_rectified_sine():
    N = 50
    x = np.sin(np.linspace(0, 2 * np.pi, N, endpoint=False))
    x = np.abs(x)
    assert sdr.crest_factor(x) == pytest.approx(np.sqrt(2), rel=1e-1)


def test_halfwave_sine():
    N = 50
    x = np.sin(np.linspace(0, 2 * np.pi, N, endpoint=False))
    x[N // 2 :] = 0
    assert sdr.crest_factor(x) == pytest.approx(2.0, rel=1e-1)


def test_square_wave():
    N = 50
    x = np.ones(N)
    x[N // 2 :] = -1
    assert sdr.crest_factor(x) == pytest.approx(1.0, rel=1e-1)


def test_pwm():
    n1 = 10
    N = 50
    x = np.zeros(N)
    x[0:n1] = 1
    assert sdr.crest_factor(x) == pytest.approx(np.sqrt(N / n1), rel=1e-1)
